"""
Contains classes and methods for examining movement events.
"""

from dataclasses import dataclass
import datetime
from enum import Enum
from typing import Dict, List, TYPE_CHECKING

from redvox.common.date_time_utils import datetime_from_epoch_microseconds_utc

if TYPE_CHECKING:
    from redvox.api1000.wrapped_redvox_packet.event_streams import Event, EventStream
    from redvox.api1000.common.mapping import Mapping


class MovementChannel(Enum):
    """
    Enumeration of movement channels.
    """
    GYROSCOPE_X: str = "GYROSCOPE_X"
    GYROSCOPE_Y: str = "GYROSCOPE_Y"
    GYROSCOPE_Z: str = "GYROSCOPE_Z"
    ACCELEROMETER_X: str = "ACCELEROMETER_X"
    ACCELEROMETER_Y: str = "ACCELEROMETER_Y"
    ACCELEROMETER_Z: str = "ACCELEROMETER_Z"


@dataclass
class MovementEvent:
    """
    Represents the metadata associated with a derived movement event.
    """
    movement_channel: MovementChannel
    movement_start: float
    movement_end: float
    movement_duration: float
    magnitude_min: float
    magnitude_max: float
    magnitude_range: float
    magnitude_mean: float
    magnitude_std_dev: float

    def movement_start_dt(self) -> datetime.datetime:
        """
        :return: The movement start as a datetime.
        """
        return datetime_from_epoch_microseconds_utc(self.movement_start)

    def movement_end_dt(self) -> datetime.datetime:
        """
        :return: The movement end as a datetime.
        """
        return datetime_from_epoch_microseconds_utc(self.movement_end)

    def movement_duration_td(self) -> datetime.timedelta:
        """
        :return: The movement duration as a timedelta.
        """
        return self.movement_end_dt() - self.movement_start_dt()

    @staticmethod
    def from_event(event: 'Event') -> 'MovementEvent':
        """
        Converts a generic API M Event into a MovementEvent.
        :param event: The Event to convert.
        :return: A MovementEvent.
        """
        numeric_payload: 'Mapping[float]' = event.get_numeric_payload()
        numeric_payload_dict: Dict[str, float] = numeric_payload.get_metadata()
        return MovementEvent(
            MovementChannel[event.get_description()],
            numeric_payload_dict["movement_start"],
            numeric_payload_dict["movement_end"],
            numeric_payload_dict["movement_duration"],
            numeric_payload_dict["magnitude_min"],
            numeric_payload_dict["magnitude_max"],
            numeric_payload_dict["magnitude_range"],
            numeric_payload_dict["magnitude_mean"],
            numeric_payload_dict["magnitude_std_dev"],
        )


@dataclass
class MovementEventStream:
    """
    Represents a derived movement event stream.
    """
    name: str
    movement_events: List[MovementEvent]

    @staticmethod
    def from_event_stream(event_stream: 'EventStream') -> 'MovementEventStream':
        """
        Converts an API M EventStream into a MovementEventStream.
        :param event_stream: Stream to convert.
        :return: A MovementEventStream.
        """
        movement_events: List[MovementEvent] = list(map(MovementEvent.from_event,
                                                        event_stream.get_events().get_values()))

        movement_events.sort(key=lambda event: event.movement_start_dt())
        return MovementEventStream(event_stream.get_name(), movement_events)

    def events_by_channel(self, movement_channel: MovementChannel) -> List[MovementEvent]:
        return list(filter(lambda event: event.movement_channel == movement_channel, self.movement_events))



"""
This module contains classes and enums for working with generic RedVox packet metadata through the cloud API.
"""
from dataclasses import dataclass
import json
from typing import Any, Callable, Dict, List, Optional, TypeVar

from dataclasses_json import dataclass_json
import requests
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM

from redvox.api1000.proto.redvox_api_m_pb2 import RedvoxPacketM
from redvox.cloud.api import ApiConfig, post_req
from redvox.cloud.routes import RoutesV1


@dataclass_json
@dataclass
class AudioMetadata:
    """
    Metadata associated with audio sensors.
    """
    sensor_name: Optional[str] = None
    sample_rate: Optional[float] = None
    first_sample_ts: Optional[int] = None
    payload_cnt: Optional[int] = None
    payload_mean: Optional[float] = None
    payload_std: Optional[float] = None
    payload_median: Optional[float] = None


@dataclass_json
@dataclass
class SingleMetadata:
    """
    Metadata associated with sensors that only contain a single dimensional channel of data
        (barometer, light, proximity).
    """
    sensor_name: Optional[str] = None
    timestamps_microseconds_utc_count: Optional[int] = None
    payload_count: Optional[int] = None
    sample_interval_mean: Optional[float] = None
    sample_interval_std: Optional[float] = None
    sample_interval_median: Optional[float] = None
    value_mean: Optional[float] = None
    value_std: Optional[float] = None
    value_median: Optional[float] = None
    metadata: Optional[List[str]] = None


@dataclass_json
@dataclass
class XyzMetadata:
    """
    Metadata for sensors that have 3-dimensions of data (accelerometer, gyroscope, magenetometer)
    """
    sensor_name: Optional[str] = None
    timestamps_microseconds_utc_count: Optional[int] = None
    payload_count: Optional[int] = None
    sample_interval_mean: Optional[float] = None
    sample_interval_std: Optional[float] = None
    sample_interval_median: Optional[float] = None
    x_mean: Optional[float] = None
    x_std: Optional[float] = None
    x_median: Optional[float] = None
    y_mean: Optional[float] = None
    y_std: Optional[float] = None
    y_median: Optional[float] = None
    z_mean: Optional[float] = None
    z_std: Optional[float] = None
    z_median: Optional[float] = None
    metadata: Optional[List[str]] = None


@dataclass_json
@dataclass
class LocationMetadata:
    """
    Metadata associated with the location sensor.
    """
    sensor_name: Optional[str] = None
    timestamps_microseconds_utc_count: Optional[int] = None
    payload_count: Optional[int] = None
    sample_interval_mean: Optional[float] = None
    sample_interval_std: Optional[float] = None
    sample_interval_median: Optional[float] = None
    latitude_mean: Optional[float] = None
    latitude_std: Optional[float] = None
    latitude_median: Optional[float] = None
    longitude_mean: Optional[float] = None
    longitude_std: Optional[float] = None
    longitude_median: Optional[float] = None
    altitude_mean: Optional[float] = None
    altitude_std: Optional[float] = None
    altitude_median: Optional[float] = None
    speed_mean: Optional[float] = None
    speed_std: Optional[float] = None
    speed_median: Optional[float] = None
    accuracy_mean: Optional[float] = None
    accuracy_std: Optional[float] = None
    accuracy_median: Optional[float] = None
    metadata: Optional[List[str]] = None


@dataclass_json
@dataclass
class PacketMetadataResult:
    """
    Metadata associated with RedVox API 900 packets.
    """
    api: Optional[int] = None
    station_id: Optional[str] = None
    station_uuid: Optional[str] = None
    auth_email: Optional[str] = None
    is_backfilled: Optional[bool] = None
    is_private: Optional[bool] = None
    is_scrambled: Optional[bool] = None
    station_make: Optional[str] = None
    station_model: Optional[str] = None
    station_os: Optional[str] = None
    station_os_version: Optional[str] = None
    station_app_version: Optional[str] = None
    battery_level: Optional[float] = None
    station_temperature: Optional[float] = None
    acquisition_url: Optional[str] = None
    synch_url: Optional[str] = None
    auth_url: Optional[str] = None
    os_ts: Optional[int] = None
    mach_ts: Optional[int] = None
    server_ts: Optional[int] = None
    data_key: Optional[str] = None
    mach_time_zero: Optional[float] = None
    best_latency: Optional[float] = None
    best_offset: Optional[float] = None
    audio_sensor: Optional[AudioMetadata] = None
    barometer_sensor: Optional[SingleMetadata] = None
    location_sensor: Optional[LocationMetadata] = None
    time_synchronization_sensor: Optional[SingleMetadata] = None
    accelerometer_sensor: Optional[XyzMetadata] = None
    magnetometer_sensor: Optional[XyzMetadata] = None
    gyroscope_sensor: Optional[XyzMetadata] = None
    light_sensor: Optional[SingleMetadata] = None
    proximity_sensor: Optional[SingleMetadata] = None


@dataclass_json
@dataclass
class AvailableMetadata:
    """
    Contains definitions for all available metadata that an be requested from the cloud api.
    """
    Api: str = "Api"
    StationId: str = "StationId"
    StationUuid: str = "StationUuid"
    AuthEmail: str = "AuthEmail"
    IsBackfilled: str = "IsBackfilled"
    IsPrivate: str = "IsPrivate"
    IsScrambled: str = "IsScrambled"
    StationMake: str = "StationMake"
    StationModel: str = "StationModel"
    StationOs: str = "StationOs"
    StationOsVersion: str = "StationOsVersion"
    StationAppVersion: str = "StationAppVersion"
    BatteryLevel: str = "BatteryLevel"
    StationTemperature: str = "StationTemperature"
    AcquisitionUrl: str = "AcquisitionUrl"
    SynchUrl: str = "SynchUrl"
    AuthUrl: str = "AuthUrl"
    OsTs: str = "OsTs"
    MachTs: str = "MachTs"
    ServerTs: str = "ServerTs"
    DataKey: str = "DataKey"
    MachTimeZero: str = "MachTimeZero"
    BestLatency: str = "BestLatency"
    BestOffset: str = "BestOffset"
    AudioSensor: str = "AudioSensor"
    BarometerSensor: str = "BarometerSensor"
    AccelerometerSensor: str = "AccelerometerSensor"
    GyroscopeSensor: str = "GyroscopeSensor"
    TimeSynchronizationSensor: str = "TimeSynchronizationSensor"
    MagnetometerSensor: str = "MagnetometerSensor"
    LightSensor: str = "LightSensor"
    ProximitySensor: str = "ProximitySensor"
    LocationSensor: str = "LocationSensor"

    @staticmethod
    def all_available_metadata() -> List[str]:
        """
        Returns a list of all available metadata definitions.
        :return: A list of all available metadata definitions.
        """
        return [
            AvailableMetadata.Api,
            AvailableMetadata.StationId,
            AvailableMetadata.StationUuid,
            AvailableMetadata.AuthEmail,
            AvailableMetadata.IsBackfilled,
            AvailableMetadata.IsPrivate,
            AvailableMetadata.IsScrambled,
            AvailableMetadata.StationMake,
            AvailableMetadata.StationModel,
            AvailableMetadata.StationOs,
            AvailableMetadata.StationOsVersion,
            AvailableMetadata.StationAppVersion,
            AvailableMetadata.BatteryLevel,
            AvailableMetadata.StationTemperature,
            AvailableMetadata.AcquisitionUrl,
            AvailableMetadata.SynchUrl,
            AvailableMetadata.AuthUrl,
            AvailableMetadata.OsTs,
            AvailableMetadata.MachTs,
            AvailableMetadata.ServerTs,
            AvailableMetadata.DataKey,
            AvailableMetadata.MachTimeZero,
            AvailableMetadata.BestLatency,
            AvailableMetadata.BestOffset,
            AvailableMetadata.AudioSensor,
            AvailableMetadata.BarometerSensor,
            AvailableMetadata.AccelerometerSensor,
            AvailableMetadata.GyroscopeSensor,
            AvailableMetadata.TimeSynchronizationSensor,
            AvailableMetadata.MagnetometerSensor,
            AvailableMetadata.LightSensor,
            AvailableMetadata.ProximitySensor,
            AvailableMetadata.LocationSensor,
        ]


@dataclass_json
@dataclass
class MetadataReq:
    """
    The definition of a metadata request. Fields should include the definitions defined by AvailableMetadata.
    """
    auth_token: str
    start_ts_s: int
    end_ts_s: int
    station_ids: List[str]
    fields: List[str]
    secret_token: Optional[str] = None


@dataclass_json
@dataclass
class MetadataResp:
    """
    Response of a metadata request.
    """
    metadata: List[PacketMetadataResult]


T = TypeVar('T')


def _get(key: str, json_dict: Dict, default: Optional[T] = None) -> Optional[T]:
    if key in json_dict:
        return json_dict[key]

    return default


@dataclass
class AdditionalMetadata:
    data_key: Optional[str]

    @staticmethod
    def from_dict(json_dict: Optional[Dict]) -> Optional['AdditionalMetadata']:
        if json_dict is None:
            return None

        return AdditionalMetadata(_get("data_key", json_dict))


@dataclass
class DbPacket:
    _id: Optional[str]
    metadata: Optional[WrappedRedvoxPacketM]
    additional_metadata: Optional[AdditionalMetadata]

    @staticmethod
    def from_dict(json_dict: Dict) -> 'DbPacket':
        return DbPacket(
            _get("_id", json_dict),
            WrappedRedvoxPacketM.from_json(json.dumps(json_dict["metadata"])),
            AdditionalMetadata.from_dict(_get("additional_metadata", json_dict))
        )


@dataclass
class MetadataRespM:
    db_packets: List[DbPacket]

    @staticmethod
    def from_json(json_dicts: Dict) -> 'MetadataRespM':

        return MetadataRespM(
            list(map(DbPacket.from_dict, json_dicts["db_packets"]))
        )


@dataclass_json
@dataclass
class TimingMetaRequest:
    """
    Request for timing metadata.
    """
    auth_token: str
    start_ts_s: int
    end_ts_s: int
    station_ids: List[str]
    secret_token: Optional[str] = None


@dataclass_json
@dataclass
class TimingMeta:
    """
    Timing metadata extracted from an individual packet.
    """
    station_id: str
    start_ts_os: float
    start_ts_mach: float
    server_ts: float
    mach_time_zero: float
    best_latency: float
    best_offset: float


@dataclass_json
@dataclass
class TimingMetaResponse:
    """
    Response of obtaining timing metadta.
    """
    items: List[TimingMeta]


def request_timing_metadata(api_config: ApiConfig,
                            timing_req: TimingMetaRequest,
                            session: Optional[requests.Session] = None,
                            timeout: Optional[float] = None) -> TimingMetaResponse:
    """
    Retrieve timing metadata.
    :param api_config: An instance of the API configuration.
    :param timing_req: An instance of a timing request.
    :param session: An (optional) session for re-using an HTTP client.
    :return: An instance of a timing response.
    """

    def handle_resp(resp) -> TimingMetaResponse:
        json: List[Dict] = resp.json()
        # noinspection Mypy
        items: List[TimingMeta] = list(map(TimingMeta.from_dict, json))
        return TimingMetaResponse(items)

    res: Optional[TimingMetaResponse] = post_req(api_config,
                                                 RoutesV1.TIMING_METADATA_REQ,
                                                 timing_req,
                                                 handle_resp,
                                                 session,
                                                 timeout)

    return res if res else TimingMetaResponse([])


def request_metadata(api_config: ApiConfig,
                     packet_metadata_req: MetadataReq,
                     session: Optional[requests.Session] = None,
                     timeout: Optional[float] = None) -> Optional[MetadataResp]:
    """
    Requests generic metadata from the cloud API.
    :param api_config: An instance of the API config.
    :param packet_metadata_req: An instance of a metadata request.
    :param session: An (optional) session for re-using an HTTP client.
    :return: A metadata response on successful call or None if there is an error.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], MetadataResp] = lambda resp: MetadataResp.from_dict(resp.json())
    return post_req(api_config,
                    RoutesV1.METADATA_REQ,
                    packet_metadata_req,
                    handle_resp,
                    session,
                    timeout)


def request_metadata_m(api_config: ApiConfig,
                       packet_metadata_req: MetadataReq,
                       session: Optional[requests.Session] = None,
                       timeout: Optional[float] = None) -> Optional[MetadataRespM]:
    """
    Requests generic metadata from the cloud API.
    :param api_config: An instance of the API config.
    :param packet_metadata_req: An instance of a metadata request.
    :param session: An (optional) session for re-using an HTTP client.
    :return: A metadata response on successful call or None if there is an error.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], MetadataRespM] = lambda resp: MetadataRespM.from_json(resp.json())
    return post_req(api_config,
                    RoutesV1.METADATA_REQ_M,
                    packet_metadata_req,
                    handle_resp,
                    session,
                    timeout)



import argparse

import sys
from typing import Optional

from PySide2.QtWidgets import QApplication, QMainWindow, QWidget

import redvox
from redvox.api1000.gui.data_explorer.data_explorer import DataExplorer


class RedvoxGui(QMainWindow):
    def __init__(self, base_dir: str, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.data_explorer: DataExplorer = DataExplorer(base_dir, self)
        self.setWindowTitle(f"RedVox GUI v{redvox.version()}")
        self.setCentralWidget(self.data_explorer)


def start_gui(base_dir: str) -> None:
    app: QApplication = QApplication(sys.argv)
    window: RedvoxGui = RedvoxGui(base_dir)
    window.showMaximized()
    app.exec_()


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser()
    parser.add_argument("base_dir")
    args = parser.parse_args()
    start_gui(args.base_dir)


if __name__ == "__main__":
    main()



"""
This module contains methods for interacting with the RedVox cloud based API.
"""
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

import requests

import redvox.cloud.errors as cloud_errors
from redvox.cloud.routes import RoutesV1


@dataclass
class ApiConfig:
    """
    Provides a configuration for the base API URL.
    """
    protocol: str
    host: str
    port: int

    def url(self, end_point: str) -> str:
        """
        Formats the API URL.
        :param end_point: Endpoint to use.
        :return: The formatted API URL.
        """
        return f"{self.protocol}://{self.host}:{self.port}{end_point}"

    @staticmethod
    def default() -> 'ApiConfig':
        return ApiConfig("https", "redvox.io", 8080)


def post_req(api_config: ApiConfig,
             route: str,
             req: Any,
             resp_transform: Callable[[requests.Response], Any],
             session: Optional[requests.Session] = None,
             timeout: Optional[float] = 10.0) -> Optional[Any]:
    url: str = api_config.url(route)
    # noinspection Mypy
    req_dict: Dict = req.to_dict()

    try:
        if session:
            resp: requests.Response = session.post(url, json=req_dict, timeout=timeout)
        else:
            resp = requests.post(url, json=req_dict, timeout=timeout)
        if resp.status_code == 200:
            # noinspection Mypy
            return resp_transform(resp)
        else:
            return None
    except requests.RequestException as e:
        raise cloud_errors.ApiConnectionError(f"Error making POST request to {url}: with body: {req_dict}: {e}")


def health_check(api_config: ApiConfig,
                 session: Optional[requests.Session] = None,
                 timeout: Optional[float] = 10.0) -> bool:
    """
    Check that the Cloud API endpoint is up.
    :param api_config: The API config.
    :param session: An (optional) session for re-using an HTTP client.
    :return: True if the endpoint is up, False otherwise.
    """
    url: str = api_config.url(RoutesV1.HEALTH_CHECK)

    if session:
        resp: requests.Response = session.get(url, timeout=timeout)
    else:
        resp = requests.get(url, timeout=timeout)

    if resp.status_code == 200:
        return True

    return False



"""
This module loads or reads sensor data from various sources
"""
import os
import glob
import numpy as np
import pandas as pd
from obspy import read
from typing import List, Dict, Optional
from redvox.api1000 import io as apim_io
from redvox.api900 import reader as api900_io
from redvox.common import file_statistics as fs, date_time_utils as dtu, timesync as ts
from redvox.common.sensor_data import SensorType, SensorData, Station, StationTiming, StationMetadata, DataPacket
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider
from redvox.api1000.wrapped_redvox_packet.sensors import xyz, single
from redvox.api1000.wrapped_redvox_packet import wrapped_packet as apim_wp
from dataclasses import dataclass


@dataclass
class StationSummary:
    """
    Contains a summary of each stations data read result.
    """
    station_id: str
    station_uuid: str
    os: str
    os_version: str
    app_version: str
    audio_sampling_rate: float
    total_duration: float
    start_dt: dtu.datetime
    end_dt: dtu.datetime

    @staticmethod
    def from_station(station: Station) -> 'StationSummary':
        total_duration: float = station.audio_sensor().data_duration_s()
        start_dt: dtu.datetime = dtu.datetime_from_epoch_microseconds_utc(station.audio_sensor().first_data_timestamp())
        end_dt: dtu.datetime = dtu.datetime_from_epoch_microseconds_utc(station.audio_sensor().last_data_timestamp())

        station_info = station.station_metadata
        audio = station.audio_sensor()
        return StationSummary(
            station_info.station_id,
            station_info.station_uuid,
            station_info.station_os if station_info.station_os is not None else "OS UNKNOWN",
            station_info.station_os_version if station_info.station_os_version is not None else "OS VERSION UNKNOWN",
            station_info.station_app_version if station_info.station_app_version is not None else "APP VERSION UNKNOWN",
            audio.sample_rate if audio is not None else float("NaN"),
            total_duration,
            start_dt,
            end_dt
        )


class ReadResult:
    """
    Stores station information after being read from files
    """
    def __init__(self,
                 station_id_uuid_to_stations: Dict[str, Station]):
        """
        :param station_id_uuid_to_stations: station_id:station_uuid -> station information
        """
        self.station_id_uuid_to_stations: Dict[str, Station] = station_id_uuid_to_stations
        self.__station_id_to_id_uuid: Dict[str, str] = {}
        self.__station_summaries: List[StationSummary] = []

        self.update_metadata()

    def update_metadata(self):
        """
        updates ids:uuids pairs and summary information
        """
        for id_uuid, station in self.station_id_uuid_to_stations.items():
            s: List[str] = id_uuid.split(":")
            self.__station_id_to_id_uuid[s[0]] = s[1]
            self.__station_summaries.append(StationSummary.from_station(station))

    def __get_station_id_by_uuid(self, uuid: str) -> str:
        """
        given a uuid, returns the station_id
        :param uuid: uuid to search for
        :return: the station_id of the uuid, or an empty string if uuid doesn't exist
        """
        for station_id, station_uuid in self.__station_id_to_id_uuid.items():
            if station_uuid == uuid:
                return station_id
        return ""

    def pop_station(self, station_id: str) -> 'ReadResult':
        """
        removes a station from the ReadResult; station_id can be one of id, uuid or id:uuid
        :param station_id: station to remove
        :return: copy of ReadResult without the station_id specified
        """
        if ":" in station_id:
            s: List[str] = station_id.split(":")
            station_id = s[0]
        elif station_id in self.__station_id_to_id_uuid.values():  # check if uuid was given
            station_id = self.__get_station_id_by_uuid(station_id)
        if self.check_for_id(station_id):
            self.station_id_uuid_to_stations.pop(f"{station_id}:{self.__station_id_to_id_uuid[station_id]}")
            self.__station_id_to_id_uuid.pop(station_id)
            summaries = self.__station_summaries.copy()  # put summaries into temp variable
            self.__station_summaries.clear()             # clear summaries and rebuild
            for summary in summaries:
                if summary.station_id != station_id:
                    self.__station_summaries.append(summary)
        else:
            print(f"ReadResult cannot remove station {station_id} because it does not exist")
        return self

    def check_for_id(self, check_id: str) -> bool:
        """
        Look at keys and shortened keys in for the check_id; must be id or uuid combination
        :param check_id: id to look for
        :return: True if check_id is in the ReadResult
        """
        return check_id in self.__station_id_to_id_uuid.keys() or check_id in self.__station_id_to_id_uuid.values()

    def get_station(self, station_id: str) -> Optional[Station]:
        """
        Find the station identified by the station_id given; it can be id or id:uuid
        :param station_id: str id of station; can be id or id:uuid
        :return: the station if it exists, None otherwise
        """
        if ":" in station_id:
            return self.station_id_uuid_to_stations[station_id]
        elif self.check_for_id(station_id):
            return self.station_id_uuid_to_stations[f"{station_id}:{self.__station_id_to_id_uuid[station_id]}"]
        print(f"WARNING: ReadResult attempted to read station id: {station_id}, but could not find id in results")
        return None

    def get_all_stations(self) -> List[Station]:
        """
        Return a list of all stations in the ReadResult
        :return: a list of all stations
        """
        return list(self.station_id_uuid_to_stations.values())

    def get_station_summaries(self) -> List[StationSummary]:
        """
        :return: A list of StationSummaries contained in this ReadResult
        """
        return self.__station_summaries

    def append_station(self, new_station_id: str, new_station: Station):
        """
        adds a station to the ReadResult.  Appends data to existing stations
        :param new_station_id: id of station to add
        :param new_station: Station object to add
        """
        if self.check_for_id(new_station_id):
            self.station_id_uuid_to_stations[new_station_id].append_station_data(new_station.station_data)
        else:
            self.station_id_uuid_to_stations[new_station_id] = new_station
            self.__station_id_to_id_uuid[new_station.station_metadata.station_id] = \
                new_station.station_metadata.station_uuid
            self.__station_summaries.append(StationSummary.from_station(new_station))

    def append(self, new_stations: 'ReadResult'):
        """
        adds stations to the ReadResult
        :param new_stations: stations to add
        """
        for new_station_id, new_station in new_stations.station_id_uuid_to_stations.items():
            self.append_station(new_station_id, new_station)


def calc_evenly_sampled_timestamps(start: float, samples: int, rate_hz: float) -> np.array:
    """
    given a start time, calculates samples amount of evenly spaced timestamps at rate_hz
    :param start: float, start timestamp
    :param samples: int, number of samples
    :param rate_hz: float, sample rate in hz
    :return: np.array with evenly spaced timestamps starting at start
    """
    return np.array(start + dtu.seconds_to_microseconds(np.arange(0, samples) / rate_hz))


def read_api900_non_mic_sensor(sensor: api900_io.RedvoxSensor, column_id: str) -> SensorData:
    """
    read a sensor that does not have mic data from an api900 data packet
    :param sensor: the non-mic api900 sensor to read
    :param column_id: string, used to name the columns
    :return: generic SensorData object
    """
    timestamps = sensor.timestamps_microseconds_utc()
    if len(timestamps) > 1:
        sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
    else:
        sample_interval = np.nan
    if type(sensor) in [api900_io.AccelerometerSensor, api900_io.MagnetometerSensor, api900_io.GyroscopeSensor]:
        data_for_df = np.transpose([timestamps,
                                    sensor.payload_values_x(), sensor.payload_values_y(), sensor.payload_values_z()])
        columns = ["timestamps", f"{column_id}_x", f"{column_id}_y", f"{column_id}_z"]
    else:
        data_for_df = np.transpose([timestamps, sensor.payload_values()])
        columns = ["timestamps", column_id]
    return SensorData(sensor.sensor_name(), pd.DataFrame(data_for_df, columns=columns), 1 / sample_interval,
                      sample_interval, False)


def read_api900_wrapped_packet(wrapped_packet: api900_io.WrappedRedvoxPacket) -> Dict[SensorType, SensorData]:
    """
    reads the data from a wrapped api900 redvox packet into a dictionary of generic data
    :param wrapped_packet: a wrapped api900 redvox packet
    :return: a dictionary containing all the sensor data
    """
    data_dict: Dict[SensorType, SensorData] = {}
    # there are 9 api900 sensors
    if wrapped_packet.has_microphone_sensor():
        sample_rate_hz = wrapped_packet.microphone_sensor().sample_rate_hz()
        timestamps = calc_evenly_sampled_timestamps(
            wrapped_packet.microphone_sensor().first_sample_timestamp_epoch_microseconds_utc(),
            fs.get_num_points_from_sample_rate(sample_rate_hz), sample_rate_hz)
        data_for_df = np.transpose([timestamps, wrapped_packet.microphone_sensor().payload_values().astype(float)])
        data_dict[SensorType.AUDIO] = SensorData(wrapped_packet.microphone_sensor().sensor_name(),
                                                 pd.DataFrame(data_for_df, columns=["timestamps", "microphone"]),
                                                 sample_rate_hz, 1 / sample_rate_hz, True)
    if wrapped_packet.has_accelerometer_sensor():
        data_dict[SensorType.ACCELEROMETER] = \
            read_api900_non_mic_sensor(wrapped_packet.accelerometer_sensor(), "accelerometer")
    if wrapped_packet.has_magnetometer_sensor():
        data_dict[SensorType.MAGNETOMETER] = \
            read_api900_non_mic_sensor(wrapped_packet.magnetometer_sensor(), "magnetometer")
    if wrapped_packet.has_gyroscope_sensor():
        data_dict[SensorType.GYROSCOPE] = read_api900_non_mic_sensor(wrapped_packet.gyroscope_sensor(), "gyroscope")
    if wrapped_packet.has_barometer_sensor():
        data_dict[SensorType.PRESSURE] = read_api900_non_mic_sensor(wrapped_packet.barometer_sensor(), "barometer")
    if wrapped_packet.has_light_sensor():
        data_dict[SensorType.LIGHT] = read_api900_non_mic_sensor(wrapped_packet.light_sensor(), "light")
    if wrapped_packet.has_infrared_sensor():
        data_dict[SensorType.INFRARED] = read_api900_non_mic_sensor(wrapped_packet.infrared_sensor(), "infrared")
    if wrapped_packet.has_image_sensor():
        data_dict[SensorType.IMAGE] = read_api900_non_mic_sensor(wrapped_packet.image_sensor(), "image")
    if wrapped_packet.has_location_sensor():
        timestamps = wrapped_packet.location_sensor().timestamps_microseconds_utc()
        if len(timestamps) > 1:
            sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
        else:
            sample_interval = np.nan
        if wrapped_packet.location_sensor().check_for_preset_lat_lon():
            lat_lon = wrapped_packet.location_sensor().get_payload_lat_lon()
            data_for_df = np.array([[timestamps[0], lat_lon[0], lat_lon[1], np.nan, np.nan, np.nan,
                                     LocationProvider.USER]])
        else:
            data_for_df = np.transpose([timestamps,
                                        wrapped_packet.location_sensor().payload_values_latitude(),
                                        wrapped_packet.location_sensor().payload_values_longitude(),
                                        wrapped_packet.location_sensor().payload_values_altitude(),
                                        wrapped_packet.location_sensor().payload_values_speed(),
                                        wrapped_packet.location_sensor().payload_values_accuracy(),
                                        np.full(len(timestamps), np.nan)])
        columns = ["timestamps", "latitude", "longitude", "altitude", "speed", "accuracy", "location_provider"]
        data_dict[SensorType.LOCATION] = SensorData(wrapped_packet.location_sensor().sensor_name(),
                                                    pd.DataFrame(data_for_df, columns=columns),
                                                    1 / sample_interval, sample_interval, False)
    return data_dict


def load_station_from_api900(api900_packet: api900_io.WrappedRedvoxPacket,
                             start_timestamp_utc_s: Optional[int] = None,
                             end_timestamp_utc_s: Optional[int] = None) -> Station:
    """
    reads in station data from a single wrapped api900 packet
    :param api900_packet: wrapped api900 packet to read from
    :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
    :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
    :return: a station Object
    """
    # set station metadata and timing
    timing = StationTiming(api900_packet.mach_time_zero(), api900_packet.microphone_sensor().sample_rate_hz(),
                           api900_packet.app_file_start_timestamp_epoch_microseconds_utc(),
                           start_timestamp_utc_s, end_timestamp_utc_s,
                           np.nan if api900_packet.best_latency() is None else api900_packet.best_latency(),
                           0.0 if api900_packet.best_offset() is None else api900_packet.best_offset())
    metadata = StationMetadata(api900_packet.redvox_id(), api900_packet.device_make(),
                               api900_packet.device_model(), False, api900_packet.device_os(),
                               api900_packet.device_os_version(), "Redvox", api900_packet.app_version(),
                               api900_packet.is_scrambled(), timing, station_uuid=api900_packet.uuid())
    data_dict = read_api900_wrapped_packet(api900_packet)
    packet_data = DataPacket(api900_packet.server_timestamp_epoch_microseconds_utc(),
                             api900_packet.app_file_start_timestamp_machine(),
                             len(api900_packet.microphone_sensor().payload_values()),
                             len(api900_packet.microphone_sensor().payload_values()) /
                             api900_packet.microphone_sensor().sample_rate_hz(),
                             float(api900_packet.start_timestamp_us_utc()), api900_packet.end_timestamp_us_utc(),
                             api900_packet.time_synchronization_sensor().payload_values(),
                             np.nan if api900_packet.best_latency() is None else api900_packet.best_latency(),
                             0.0 if api900_packet.best_offset() is None else api900_packet.best_offset())
    packet_list: List[DataPacket] = [packet_data]
    # get the best timing values for the station
    if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
        ts_analysis = ts.TimeSyncData(packet_data, metadata)
        timing.station_best_latency = ts_analysis.best_latency
        timing.station_best_offset = ts_analysis.best_offset
    return Station(metadata, data_dict, packet_list)


def load_station_from_api900_file(directory: str, start_timestamp_utc_s: Optional[int] = None,
                                  end_timestamp_utc_s: Optional[int] = None) -> Station:
    """
    reads in station data from a single api900 file
    :param directory: string of the file to read from
    :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
    :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
    :return: a station Object
    """
    api900_packet = api900_io.read_rdvxz_file(directory)
    return load_station_from_api900(api900_packet, start_timestamp_utc_s, end_timestamp_utc_s)


def load_file_range_from_api900(directory: str,
                                start_timestamp_utc_s: Optional[int] = None,
                                end_timestamp_utc_s: Optional[int] = None,
                                redvox_ids: Optional[List[str]] = None,
                                structured_layout: bool = False,
                                concat_continuous_segments: bool = True) -> ReadResult:
    """
    reads in api900 data from a directory and returns a list of stations
    note that the param descriptions are taken directly from api900.reader.read_rdvxz_file_range
    :param directory: The root directory of the data. If structured_layout is False, then this directory will
                      contain various unorganized .rdvxz files. If structured_layout is True, then this directory
                      must be the root api900 directory of the structured files.
    :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
    :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
    :param redvox_ids: An optional list of redvox_ids to filter against (default=[]).
    :param structured_layout: An optional value to define if this is loading structured data (default=False).
    :param concat_continuous_segments: An optional value to define if this function should concatenate rdvxz files
                                       into multiple continuous rdvxz files separated at gaps.
    :return: a list of Station objects that contain the data
    """
    all_stations: ReadResult = ReadResult({})
    all_data = api900_io.read_rdvxz_file_range(directory, start_timestamp_utc_s, end_timestamp_utc_s, redvox_ids,
                                               structured_layout, concat_continuous_segments)
    for redvox_id, wrapped_packets in all_data.items():
        # set station metadata and timing based on first packet
        timing = StationTiming(wrapped_packets[0].mach_time_zero(),
                               wrapped_packets[0].microphone_sensor().sample_rate_hz(),
                               wrapped_packets[0].app_file_start_timestamp_epoch_microseconds_utc(),
                               start_timestamp_utc_s, end_timestamp_utc_s,
                               np.nan if wrapped_packets[0].best_latency() is None else
                               wrapped_packets[0].best_latency(),
                               0.0 if wrapped_packets[0].best_offset() is None else wrapped_packets[0].best_offset())
        metadata = StationMetadata(wrapped_packets[0].redvox_id(), wrapped_packets[0].device_make(),
                                   wrapped_packets[0].device_model(), False, wrapped_packets[0].device_os(),
                                   wrapped_packets[0].device_os_version(), "Redvox", wrapped_packets[0].app_version(),
                                   wrapped_packets[0].is_scrambled(), timing, station_uuid=wrapped_packets[0].uuid())
        # add data from packets
        new_station = Station(metadata)
        packet_list: List[DataPacket] = []
        for packet in wrapped_packets:
            if packet.has_time_synchronization_sensor():
                time_sync = packet.time_synchronization_sensor().payload_values()
            else:
                time_sync = None
            data_dict = read_api900_wrapped_packet(packet)
            new_station.append_station_data(data_dict)
            packet_data = DataPacket(packet.server_timestamp_epoch_microseconds_utc(),
                                     packet.app_file_start_timestamp_machine(),
                                     data_dict[SensorType.AUDIO].num_samples(),
                                     data_dict[SensorType.AUDIO].data_duration_s(),
                                     float(packet.start_timestamp_us_utc()), packet.end_timestamp_us_utc(),
                                     time_sync, np.nan if packet.best_latency() is None else packet.best_latency(),
                                     0.0 if packet.best_offset() is None else packet.best_offset())
            packet_list.append(packet_data)
        new_station.packet_data = packet_list

        if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
            # get the best timing values for the station
            ts_analysis = ts.TimeSyncAnalysis(new_station)
            new_station.station_metadata.timing_data.station_best_latency = ts_analysis.get_best_latency()
            new_station.station_metadata.timing_data.station_best_offset = ts_analysis.get_best_offset()

        # create the Station data object
        all_stations.append_station(redvox_id, new_station)

    return all_stations


def read_apim_xyz_sensor(sensor: xyz.Xyz, column_id: str) -> SensorData:
    """
    read a sensor that has xyz data channels from an api M data packet
    :param sensor: the xyz api M sensor to read
    :param column_id: string, used to name the columns
    :return: generic SensorData object
    """
    timestamps = sensor.get_timestamps().get_timestamps()
    if len(timestamps) > 1:
        sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
    else:
        sample_interval = np.nan
    data_for_df = np.transpose([timestamps,
                                sensor.get_x_samples().get_values(),
                                sensor.get_y_samples().get_values(),
                                sensor.get_z_samples().get_values()])
    columns = ["timestamps", f"{column_id}_x", f"{column_id}_y", f"{column_id}_z"]
    return SensorData(sensor.get_sensor_description(), pd.DataFrame(data_for_df, columns=columns),
                      1 / sample_interval, sample_interval, False)


def read_apim_single_sensor(sensor: single.Single, column_id: str) -> SensorData:
    """
    read a sensor that has a single data channel from an api M data packet
    :param sensor: the single channel api M sensor to read
    :param column_id: string, used to name the columns
    :return: generic SensorData object
    """
    timestamps = sensor.get_timestamps().get_timestamps()
    if len(timestamps) > 1:
        sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
    else:
        sample_interval = np.nan
    data_for_df = np.transpose([timestamps, sensor.get_samples().get_values()])
    columns = ["timestamps", column_id]
    return SensorData(sensor.get_sensor_description(), pd.DataFrame(data_for_df, columns=columns),
                      sample_interval, sample_interval, False)


def load_apim_wrapped_packet(wrapped_packet: apim_wp.WrappedRedvoxPacketM) -> Dict[SensorType, SensorData]:
    """
    reads the data from a wrapped api M redvox packet into a dictionary of generic data
    :param wrapped_packet: a wrapped api M redvox packet
    :return: a dictionary containing all the sensor data
    """
    data_dict: Dict[SensorType, SensorData] = {}
    sensors = wrapped_packet.get_sensors()
    # there are 16 api M sensors
    if sensors.has_audio() and sensors.validate_audio():
        sample_rate_hz = sensors.get_audio().get_sample_rate()
        data_for_df = sensors.get_audio().get_samples().get_values()
        timestamps = calc_evenly_sampled_timestamps(sensors.get_audio().get_first_sample_timestamp(),
                                                    len(data_for_df), sample_rate_hz)
        data_dict[SensorType.AUDIO] = SensorData(sensors.get_audio().get_sensor_description(),
                                                 pd.DataFrame(np.transpose([timestamps, data_for_df]),
                                                              columns=["timestamps", "microphone"]),
                                                 sample_rate_hz, 1 / sample_rate_hz, True)
    if sensors.has_compress_audio() and sensors.validate_compressed_audio():
        sample_rate_hz = sensors.get_compressed_audio().get_sample_rate()
        data_for_df = sensors.get_compressed_audio().get_samples().get_values()
        timestamps = calc_evenly_sampled_timestamps(sensors.get_compressed_audio().get_first_sample_timestamp(),
                                                    len(data_for_df), sample_rate_hz)
        data_dict[SensorType.COMPRESSED_AUDIO] = SensorData(sensors.get_compressed_audio().get_sensor_description(),
                                                            pd.DataFrame(np.transpose([timestamps, data_for_df]),
                                                                         columns=["compressed_audio"]),
                                                            sample_rate_hz, 1 / sample_rate_hz, True)
    if sensors.has_accelerometer() and sensors.validate_accelerometer():
        data_dict[SensorType.ACCELEROMETER] = read_apim_xyz_sensor(sensors.get_accelerometer(), "accelerometer")
    if sensors.has_magnetometer() and sensors.validate_magnetometer():
        data_dict[SensorType.MAGNETOMETER] = read_apim_xyz_sensor(sensors.get_magnetometer(), "magnetometer")
    if sensors.has_linear_acceleration() and sensors.validate_accelerometer():
        data_dict[SensorType.LINEAR_ACCELERATION] = read_apim_xyz_sensor(sensors.get_linear_acceleration(),
                                                                         "linear_accel")
    if sensors.has_orientation() and sensors.validate_orientation():
        data_dict[SensorType.ORIENTATION] = read_apim_xyz_sensor(sensors.get_orientation(), "orientation")
    if sensors.has_rotation_vector() and sensors.validate_rotation_vector():
        data_dict[SensorType.ROTATION_VECTOR] = read_apim_xyz_sensor(sensors.get_rotation_vector(), "rotation_vector")
    if sensors.has_gyroscope() and sensors.validate_gyroscope():
        data_dict[SensorType.GYROSCOPE] = read_apim_xyz_sensor(sensors.get_gyroscope(), "gyroscope")
    if sensors.has_gravity() and sensors.validate_gravity():
        data_dict[SensorType.GRAVITY] = read_apim_xyz_sensor(sensors.get_gravity(), "gravity")
    if sensors.has_pressure() and sensors.validate_pressure():
        data_dict[SensorType.PRESSURE] = read_apim_single_sensor(sensors.get_pressure(), "barometer")
    if sensors.has_light() and sensors.validate_light():
        data_dict[SensorType.LIGHT] = read_apim_single_sensor(sensors.get_light(), "light")
    if sensors.has_proximity() and sensors.validate_proximity():
        data_dict[SensorType.PROXIMITY] = read_apim_single_sensor(sensors.get_proximity(), "proximity")
    if sensors.has_ambient_temperature() and sensors.validate_ambient_temperature():
        data_dict[SensorType.TEMPERATURE] = read_apim_single_sensor(sensors.get_ambient_temperature(), "ambient_temp")
    if sensors.has_relative_humidity() and sensors.validate_relative_humidity():
        data_dict[SensorType.RELATIVE_HUMIDITY] = read_apim_single_sensor(sensors.get_relative_humidity(),
                                                                          "rel_humidity")
    if sensors.has_image() and sensors.validate_image():
        timestamps = sensors.get_image().get_timestamps().get_timestamps()
        if len(timestamps) > 1:
            sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
        else:
            sample_interval = np.nan
        data_for_df = np.transpose([timestamps, sensors.get_image().get_samples()])
        data_dict[SensorType.IMAGE] = SensorData(sensors.get_image().get_sensor_description(),
                                                 pd.DataFrame(data_for_df, columns=["image"]),
                                                 1 / sample_interval, sample_interval, False)
    if sensors.has_location():
        if sensors.validate_location():
            timestamps = sensors.get_location().get_timestamps().get_timestamps()
            if len(timestamps) > 1:
                sample_interval = dtu.microseconds_to_seconds(float(np.mean(np.diff(timestamps))))
            else:
                sample_interval = np.nan
            data_for_df = np.transpose([timestamps,
                                        sensors.get_location().get_latitude_samples().get_values(),
                                        sensors.get_location().get_longitude_samples().get_values(),
                                        sensors.get_location().get_altitude_samples().get_values(),
                                        sensors.get_location().get_speed_samples().get_values(),
                                        sensors.get_location().get_bearing_samples().get_values(),
                                        sensors.get_location().get_horizontal_accuracy_samples().get_values(),
                                        sensors.get_location().get_vertical_accuracy_samples().get_values(),
                                        sensors.get_location().get_speed_samples().get_values(),
                                        sensors.get_location().get_bearing_accuracy_samples().get_values(),
                                        sensors.get_location().get_location_providers().get_values()])
        elif sensors.get_location().get_last_best_location():
            timestamps = [sensors.get_location().get_last_best_location().get_latitude_longitude_timestamp().get_mach()]
            sample_interval = np.nan
            data_for_df = np.transpose([[timestamps],
                                        [sensors.get_location().get_last_best_location().get_latitude()],
                                        [sensors.get_location().get_last_best_location().get_longitude()],
                                        [sensors.get_location().get_last_best_location().get_altitude()],
                                        [sensors.get_location().get_last_best_location().get_speed()],
                                        [sensors.get_location().get_last_best_location().get_bearing()],
                                        [sensors.get_location().get_last_best_location().get_horizontal_accuracy()],
                                        [sensors.get_location().get_last_best_location().get_vertical_accuracy()],
                                        [sensors.get_location().get_last_best_location().get_speed_accuracy()],
                                        [sensors.get_location().get_last_best_location().get_bearing_accuracy()],
                                        [sensors.get_location().get_last_best_location().get_location_provider()]])
        else:
            # well, there's no location, so there's nothing left to do but
            return data_dict
        # if here, location was good, add it in
        columns = ["timestamps", "latitude", "longitude", "altitude", "speed", "bearing",
                   "horizontal_accuracy", "vertical_accuracy", "speed_accuracy", "bearing_accuracy",
                   "location_provider"]
        data_dict[SensorType.LOCATION] = SensorData(sensors.get_location().get_sensor_description(),
                                                    pd.DataFrame(data_for_df, columns=columns),
                                                    1 / sample_interval, sample_interval, False)
    return data_dict


def load_station_from_apim(directory: str, start_timestamp_utc_s: Optional[int] = None,
                           end_timestamp_utc_s: Optional[int] = None) -> Station:
    """
    reads in station data from a single api M file
    :param directory: string of the file to read from
    :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
    :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
    :return: a station Object
    """
    read_packet = apim_io.read_rdvxm_file(directory)
    # set station metadata and timing based on first packet
    if read_packet.get_sensors().validate_audio():
        timing = StationTiming(read_packet.get_timing_information().get_app_start_mach_timestamp(),
                               read_packet.get_sensors().get_audio().get_sample_rate(),
                               read_packet.get_sensors().get_audio().get_first_sample_timestamp(),
                               start_timestamp_utc_s, end_timestamp_utc_s,
                               np.nan if read_packet.get_timing_information().get_best_latency() is None else
                               read_packet.get_timing_information().get_best_latency(),
                               0.0 if read_packet.get_timing_information().get_best_offset() is None else
                               read_packet.get_timing_information().get_best_offset())
    else:
        raise ValueError("Station is missing Audio sensor!")
    metadata = StationMetadata(read_packet.get_station_information().get_id(),
                               read_packet.get_station_information().get_make(),
                               read_packet.get_station_information().get_model(), False,
                               read_packet.get_station_information().get_os().name,
                               read_packet.get_station_information().get_os_version(), "Redvox",
                               read_packet.get_station_information().get_app_version(),
                               read_packet.get_station_information().get_app_settings().get_scramble_audio_data(),
                               timing, station_uuid=read_packet.get_station_information().get_uuid())
    # add data from packets
    time_sync_exchanges = read_packet.get_timing_information().get_synch_exchanges().get_values()
    time_sync = []
    for exchange in time_sync_exchanges:
        time_sync.extend([exchange.get_a1(), exchange.get_a2(), exchange.get_a3(),
                          exchange.get_b1(), exchange.get_b2(), exchange.get_b3()])
    data_dict = load_apim_wrapped_packet(read_packet)
    packet_data = DataPacket(read_packet.get_timing_information().get_server_acquisition_arrival_timestamp(),
                             read_packet.get_timing_information().get_app_start_mach_timestamp(),
                             data_dict[SensorType.AUDIO].num_samples(),
                             data_dict[SensorType.AUDIO].data_duration_s(),
                             read_packet.get_timing_information().get_packet_start_mach_timestamp(),
                             read_packet.get_timing_information().get_packet_end_mach_timestamp(),
                             np.array(time_sync),
                             np.nan if read_packet.get_timing_information().get_best_latency() is None else
                             read_packet.get_timing_information().get_best_latency(),
                             0.0 if read_packet.get_timing_information().get_best_offset() is None else
                             read_packet.get_timing_information().get_best_offset())
    packet_list: List[DataPacket] = [packet_data]
    # get the best timing values for the station
    if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
        ts_analysis = ts.TimeSyncData(packet_data, metadata)
        timing.station_best_latency = ts_analysis.best_latency
        timing.station_best_offset = ts_analysis.best_offset
    return Station(metadata, data_dict, packet_list)


def load_from_file_range_api_m(directory: str,
                               start_timestamp_utc_s: Optional[int] = None,
                               end_timestamp_utc_s: Optional[int] = None,
                               redvox_ids: Optional[List[str]] = None,
                               structured_layout: bool = False) -> ReadResult:
    """
    reads in api M data from a directory and returns a list of stations
    :param directory: The root directory of the data. If structured_layout is False, then this directory will
                      contain various unorganized .rdvxm files. If structured_layout is True, then this directory
                      must be the root api1000 directory of the structured files.
    :param start_timestamp_utc_s: The start timestamp in seconds since the epoch UTC.
    :param end_timestamp_utc_s: The end timestamp in seconds since the epoch UTC.
    :param redvox_ids: An optional list of redvox_ids to filter against, default empty list
    :param structured_layout: An optional value to define if this is loading structured data, default False.
    :return: a list of Station objects that contain the data
    """
    all_stations: ReadResult = ReadResult({})
    all_data = apim_io.read_structured(directory, start_timestamp_utc_s, end_timestamp_utc_s, redvox_ids,
                                       structured_layout)
    for read_packets in all_data.all_wrapped_packets:
        # set station metadata and timing based on first packet
        if read_packets.wrapped_packets[0]:
            if read_packets.wrapped_packets[0].get_sensors().get_audio():
                first_pack = read_packets.wrapped_packets[0]
                timing = StationTiming(read_packets.start_mach_timestamp, read_packets.audio_sample_rate,
                                       first_pack.get_sensors().get_audio().get_first_sample_timestamp(),
                                       start_timestamp_utc_s, end_timestamp_utc_s,
                                       np.nan if first_pack.get_timing_information().get_best_latency() is None else
                                       first_pack.get_timing_information().get_best_latency(),
                                       0.0 if first_pack.get_timing_information().get_best_offset() is None else
                                       first_pack.get_timing_information().get_best_offset())
                station_info = first_pack.get_station_information()
                metadata = StationMetadata(read_packets.redvox_id, station_info.get_make(), station_info.get_model(),
                                           False, station_info.get_os().name, station_info.get_os_version(),
                                           "Redvox", station_info.get_app_version(),
                                           station_info.get_app_settings().get_scramble_audio_data(), timing,
                                           station_uuid=read_packets.uuid)
            else:
                raise ValueError("Error reading data window: Packet is missing Audio sensor!")
        else:
            raise ValueError("Error reading data window: First packet of data is missing!")
        new_station = Station(metadata)
        # add data from packets
        packet_list: List[DataPacket] = []
        for packet in read_packets.wrapped_packets:
            time_sync = packet.get_timing_information().get_synch_exchange_array()
            data_dict = load_apim_wrapped_packet(packet)
            new_station.append_station_data(data_dict)
            packet_data = DataPacket(packet.get_timing_information().get_server_acquisition_arrival_timestamp(),
                                     packet.get_timing_information().get_app_start_mach_timestamp(),
                                     data_dict[SensorType.AUDIO].num_samples(),
                                     data_dict[SensorType.AUDIO].data_duration_s(),
                                     packet.get_timing_information().get_packet_start_mach_timestamp(),
                                     packet.get_timing_information().get_packet_end_mach_timestamp(),
                                     np.array(time_sync),
                                     np.nan if packet.get_timing_information().get_best_latency() is None else
                                     packet.get_timing_information().get_best_latency(),
                                     0.0 if packet.get_timing_information().get_best_offset() is None else
                                     packet.get_timing_information().get_best_offset())
            packet_list.append(packet_data)
        new_station.packet_data = packet_list

        if timing.station_best_latency is None or np.isnan(timing.station_best_latency):
            # get the best timing values for the station
            ts_analysis = ts.TimeSyncAnalysis(new_station)
            new_station.station_metadata.timing_data.station_best_latency = ts_analysis.get_best_latency()
            new_station.station_metadata.timing_data.station_best_offset = ts_analysis.get_best_offset()

        # create the Station data object
        all_stations.append_station(f"{read_packets.redvox_id}:{read_packets.uuid}", new_station)
    return all_stations


def load_from_mseed(file_path: str, station_ids: Optional[List[str]] = None) -> ReadResult:
    """
    load station data from a miniseed file
    :param file_path: the location of the miniseed file
    :param station_ids: the station ids to search for, default None; if None, get all stations
    :return: a list of Station objects that contain the data
    """
    stations: ReadResult = ReadResult({})
    strm = read(file_path)
    for data_stream in strm:
        record_info = data_stream.meta
        start_time = int(dtu.seconds_to_microseconds(data_stream.meta["starttime"].timestamp))
        end_time = int(dtu.seconds_to_microseconds(data_stream.meta["endtime"].timestamp))
        station_timing = StationTiming(np.nan, record_info["sampling_rate"], start_time, start_time, end_time)
        station_id = f'{record_info["network"]}{record_info["station"]}_{record_info["location"]}'
        metadata = StationMetadata(station_id, "mb3_make", "mb3_model", False, "mb3_os", "mb3_os_vers",
                                   "mb3_recorder", "mb3_recorder_version", False, station_timing,
                                   record_info["calib"], record_info["network"], record_info["station"],
                                   record_info["location"], record_info["channel"], record_info["mseed"]["encoding"])
        sample_rate_hz = record_info["sampling_rate"]
        timestamps = calc_evenly_sampled_timestamps(start_time, int(record_info["npts"]), sample_rate_hz)
        data_for_df = np.transpose([timestamps, data_stream.data])
        sensor_data = SensorData(record_info["channel"], pd.DataFrame(data_for_df, columns=["timestamps", "BDF"]),
                                 record_info["sampling_rate"], True)
        data_packet = DataPacket(np.nan, start_time, start_time, end_time)
        if station_ids is None or station_id in station_ids:
            stations.append_station(f"{station_id}:{station_id}", Station(metadata, {SensorType.AUDIO: sensor_data},
                                                                          [data_packet]))
    return stations


def read_all_in_dir(directory: str,
                    start_timestamp_utc_s: Optional[int] = None,
                    end_timestamp_utc_s: Optional[int] = None,
                    redvox_ids: Optional[List[str]] = None,
                    structured_layout: bool = False) -> ReadResult:
    """
    load all data files in the directory
    :param directory: location of all the files; if structured_layout is True, the directory contains a root api1000,
                        api900, or mseed directory, if structured_layout is False, the directory contains unsorted files
    :param start_timestamp_utc_s: The start timestamp as seconds since the epoch UTC.
    :param end_timestamp_utc_s: The end timestamp as seconds since the epoch UTC.
    :param redvox_ids: An optional list of redvox_ids to filter against, default empty list
    :param structured_layout: An optional value to define if this is loading structured data, default False.
    :return: a list of Station objects that contain the data
    """
    # create the object to store the data
    stations: ReadResult = ReadResult({})
    # if structured_layout, there should be a specifically named folder in directory
    if structured_layout:
        if "api900" not in directory:
            api900_dir = os.path.join(directory, "api900")
        else:
            api900_dir = directory
        if "api1000" not in directory:
            apim_dir = os.path.join(directory, "api1000")
        else:
            apim_dir = directory
        if "mseed" not in directory:
            mseed_dir = os.path.join(directory, "mseed")
        else:
            mseed_dir = directory
        # check if none of the paths exists
        if not (os.path.exists(api900_dir) or os.path.exists(apim_dir) or os.path.exists(mseed_dir)):
            # no specially named directory found; raise error
            raise ValueError(f"{directory} does not contain api900, api1000 or mseed directory.")
    else:
        # load files from unstructured layout; everything is sitting in the main directory
        api900_dir = directory
        apim_dir = directory
        mseed_dir = directory

    # get api900 data
    stations.append(load_file_range_from_api900(api900_dir, start_timestamp_utc_s, end_timestamp_utc_s,
                                                redvox_ids, structured_layout, False))
    # get api1000 data
    stations.append(load_from_file_range_api_m(apim_dir, start_timestamp_utc_s, end_timestamp_utc_s,
                                               redvox_ids, structured_layout))
    # get mseed data
    all_paths = glob.glob(os.path.join(mseed_dir, "*.mseed"))
    for path in all_paths:
        stations.append(load_from_mseed(path, redvox_ids))
    return stations



from datetime import datetime
from typing import List, Optional

from PySide2.QtWidgets import QTreeWidgetItem, QTreeWidget, QWidget

from redvox.api1000.common.common import SamplePayload, SummaryStatistics
from redvox.api1000.common.metadata import Metadata
from redvox.api1000.gui.data_explorer.packet_details import DateTimes
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM


def make_metadata_item(metadata: Metadata) -> QTreeWidgetItem:
    metadata_item: QTreeWidgetItem = QTreeWidgetItem(["metadata", f"<cnt {metadata.get_metadata_count()}>"])
    for key, value in metadata.get_metadata().items():
        metadata_item.addChild(QTreeWidgetItem([key, value]))
    return metadata_item


def make_summary_statistics(summary_statistics: SummaryStatistics, name: str = "value_statistics") -> QTreeWidgetItem:
    summary_stats_item: QTreeWidgetItem = QTreeWidgetItem([name])
    summary_stats_item.addChildren([
        QTreeWidgetItem(["count", str(summary_statistics.get_count())]),
        QTreeWidgetItem(["mean", str(summary_statistics.get_mean())]),
        QTreeWidgetItem(["standard_deviation", str(summary_statistics.get_standard_deviation())]),
        QTreeWidgetItem(["min", str(summary_statistics.get_min())]),
        QTreeWidgetItem(["max", str(summary_statistics.get_max())]),
        QTreeWidgetItem(["range", str(summary_statistics.get_range())]),
        make_metadata_item(summary_statistics.get_metadata())
    ])
    return summary_stats_item


class PacketExplorer(QTreeWidget):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setColumnCount(2)
        self.setHeaderLabels([
            "Field",
            "Value"
        ])

        self.wrapped_packet: Optional[WrappedRedvoxPacketM] = None

    def make_values_item(self, values_type: str, values_cnt: int) -> QTreeWidgetItem:
        values_item: QTreeWidgetItem = QTreeWidgetItem([f"<{values_type} values>", f"<cnt {values_cnt}>"])
        return values_item

    def make_sample_payload_item(self, sensor_type: str, sample_payload: SamplePayload) -> QTreeWidgetItem:
        sample_payload_item: QTreeWidgetItem = QTreeWidgetItem([f"<{sensor_type}> sample_payload"])
        sample_payload_item.addChildren([
            QTreeWidgetItem(["unit", sample_payload.get_unit().name]),
            self.make_values_item(sensor_type, sample_payload.get_values_count()),
            make_summary_statistics(sample_payload.get_summary_statistics(), "value_statistics"),
            make_metadata_item(sample_payload.get_metadata())
        ])
        return sample_payload_item

    def check_if_value_item(self, current: QTreeWidgetItem, previous: Optional[QTreeWidgetItem]):
        text: str = current.text(0)
        if " values>" in text:
            values: List[float] = []
            if text == "<audio values>":
                audio = self.wrapped_packet.get_sensors().get_audio()
                values = list(audio.get_samples().get_values())
                dts = DateTimes.from_sr(audio.get_first_sample_timestamp(),
                                        audio.get_sample_rate(),
                                        audio.get_samples().get_values_count())
                self.parentWidget().parentWidget().details_column.values_widget.update_from_values(values, dts)
                self.parentWidget().parentWidget().details_column.plot_widget.update_from_values(values, dts)

    def update_from_packet(self, wrapped_packet: WrappedRedvoxPacketM) -> None:
        self.wrapped_packet = wrapped_packet
        self.clear()

        station_info = wrapped_packet.get_station_information()

        app_settings_item: QTreeWidgetItem = QTreeWidgetItem(["app_settings"])
        app_settings = station_info.get_app_settings()
        app_settings_item.addChildren([
            QTreeWidgetItem(["audio_sampling_rate", app_settings.get_audio_sampling_rate().name]),
            QTreeWidgetItem(["audio_source_tuning", app_settings.get_audio_source_tuning().name]),
            QTreeWidgetItem(
                ["additional_input_sensors", str(app_settings.get_additional_input_sensors().get_values())]),
            QTreeWidgetItem(["automatically_record", str(app_settings.get_automatically_record())]),
            QTreeWidgetItem(["launch_at_power_up", str(app_settings.get_launch_at_power_up())]),
            QTreeWidgetItem(["station_id", app_settings.get_station_id()]),
            QTreeWidgetItem(["push_to_server", str(app_settings.get_push_to_server())]),
            QTreeWidgetItem(["publish_data_as_private", str(app_settings.get_publish_data_as_private())]),
            QTreeWidgetItem(["scramble_audio_data", str(app_settings.get_scramble_audio_data())]),
            QTreeWidgetItem(["provide_backfill", str(app_settings.get_provide_backfill())]),
            QTreeWidgetItem(["remove_sensor_dc_offset", str(app_settings.get_remove_sensor_dc_offset())]),
            QTreeWidgetItem(["fft_overlap", str(app_settings.get_fft_overlap().name)]),
            QTreeWidgetItem(["use_custom_time_sync_server", str(app_settings.get_use_custom_time_sync_server())]),
            QTreeWidgetItem(["time_sync_server_url", app_settings.get_time_sync_server_url()]),
            QTreeWidgetItem(["use_custom_data_server", str(app_settings.get_use_custom_data_server())]),
            QTreeWidgetItem(["data_server_url", app_settings.get_data_server_url()]),
            QTreeWidgetItem(["auto_delete_data_files", str(app_settings.get_auto_delete_data_files())]),
            QTreeWidgetItem(["storage_space_allowance", str(app_settings.get_storage_space_allowance())]),
            QTreeWidgetItem(["use_sd_card_for_data_storage", str(app_settings.get_use_sd_card_for_data_storage())]),
            QTreeWidgetItem(["use_location_services", str(app_settings.get_use_location_services())]),
            QTreeWidgetItem(["use_latitude", str(app_settings.get_use_latitude())]),
            QTreeWidgetItem(["use_longitude", str(app_settings.get_use_longitude())]),
            QTreeWidgetItem(["use_altitude", str(app_settings.get_use_altitude())]),
            make_metadata_item(app_settings.get_metadata()),
        ])

        station_metrics_item: QTreeWidgetItem = QTreeWidgetItem(["station_metrics"])
        station_metrics = station_info.get_station_metrics()
        station_metrics_item.addChildren([
            self.make_sample_payload_item("network_strength", station_metrics.get_network_strength()),
            self.make_sample_payload_item("temperature", station_metrics.get_temperature()),
            self.make_sample_payload_item("battery", station_metrics.get_battery()),
            self.make_sample_payload_item("battery_current", station_metrics.get_battery_current()),
            self.make_sample_payload_item("available_ram", station_metrics.get_available_ram()),
            self.make_sample_payload_item("available_disk", station_metrics.get_available_disk()),
            self.make_sample_payload_item("cpu_utilization", station_metrics.get_cpu_utilization()),
        ])

        service_urls_item: QTreeWidgetItem = QTreeWidgetItem(["service_urls"])
        service_urls = station_info.get_service_urls()
        service_urls_item.addChildren([
            QTreeWidgetItem(["auth_server", service_urls.get_auth_server()]),
            QTreeWidgetItem(["synch_server", service_urls.get_synch_server()]),
            QTreeWidgetItem(["acquisition_server", service_urls.get_acquisition_server()]),
            make_metadata_item(service_urls.get_metadata())
        ])

        station_info_item: QTreeWidgetItem = QTreeWidgetItem(["station_information"])
        station_info_item.addChildren([
            QTreeWidgetItem(["id", station_info.get_id()]),
            QTreeWidgetItem(["uuid", station_info.get_uuid()]),
            QTreeWidgetItem(["description", station_info.get_description()]),
            QTreeWidgetItem(["auth_id", station_info.get_auth_id()]),
            QTreeWidgetItem(["make", station_info.get_make()]),
            QTreeWidgetItem(["model", station_info.get_model()]),
            QTreeWidgetItem(["os", station_info.get_os().name]),
            QTreeWidgetItem(["os_version", station_info.get_os_version()]),
            QTreeWidgetItem(["app_version", station_info.get_app_version()]),
            QTreeWidgetItem(["is_private", str(station_info.get_is_private())]),
            app_settings_item,
            station_metrics_item,
            service_urls_item,
            make_metadata_item(station_info.get_metadata())
        ])

        timing_info_item: QTreeWidgetItem = QTreeWidgetItem(["timing_information"])
        timing_info = wrapped_packet.get_timing_information()
        synch_exchanges_item: QTreeWidgetItem = QTreeWidgetItem(
            ["synch_exchanges", f"<cnt {timing_info.get_synch_exchanges().get_count()}>"])

        for exchange in timing_info.get_synch_exchanges().get_values():
            synch_exchanges_item.addChildren([
                QTreeWidgetItem(["a1", str(exchange.get_a1())]),
                QTreeWidgetItem(["a2", str(exchange.get_a2())]),
                QTreeWidgetItem(["a3", str(exchange.get_a3())]),
                QTreeWidgetItem(["b1", str(exchange.get_a1())]),
                QTreeWidgetItem(["b2", str(exchange.get_a2())]),
                QTreeWidgetItem(["b3", str(exchange.get_a3())]),
                QTreeWidgetItem(["unit", str(exchange.get_unit())]),
                make_metadata_item(exchange.get_metadata())
            ])

        timing_info_item.addChildren([
            QTreeWidgetItem(["packet_start_os_timestamp", str(timing_info.get_packet_start_os_timestamp())]),
            QTreeWidgetItem(["packet_start_mach_timestamp", str(timing_info.get_packet_start_mach_timestamp())]),
            QTreeWidgetItem(["packet_end_os_timestamp", str(timing_info.get_packet_end_os_timestamp())]),
            QTreeWidgetItem(["packet_end_mach_timestamp", str(timing_info.get_packet_end_mach_timestamp())]),
            QTreeWidgetItem(["server_acquisition_arrival_timestamp",
                             str(timing_info.get_server_acquisition_arrival_timestamp())]),
            QTreeWidgetItem(["app_start_mach_timestamp", str(timing_info.get_app_start_mach_timestamp())]),
            synch_exchanges_item,
            QTreeWidgetItem(["best_latency", str(timing_info.get_best_latency())]),
            QTreeWidgetItem(["best_offset", str(timing_info.get_best_offset())]),
            QTreeWidgetItem(["score", str(timing_info.get_score())]),
            QTreeWidgetItem(["score_method", str(timing_info.get_score_method().name)]),
            QTreeWidgetItem(["unit", str(timing_info.get_unit().name)]),
            make_metadata_item(timing_info.get_metadata()),
        ])

        sensors_item: QTreeWidgetItem = QTreeWidgetItem(["sensors"])
        sensors = wrapped_packet.get_sensors()
        audio_item: QTreeWidgetItem = QTreeWidgetItem(["audio"])
        audio = sensors.get_audio()

        if audio:
            audio_item.addChildren([
                QTreeWidgetItem(["sensor_description", str(audio.get_sensor_description())]),
                QTreeWidgetItem(["first_sample_timestamp", str(audio.get_first_sample_timestamp())]),
                QTreeWidgetItem(["sample_rate", str(audio.get_sample_rate())]),
                QTreeWidgetItem(["bits_of_precision", str(audio.get_bits_of_precision())]),
                QTreeWidgetItem(["is_scrambled", str(audio.get_is_scrambled())]),
                QTreeWidgetItem(["encoding", str(audio.get_encoding())]),
                self.make_sample_payload_item("audio", audio.get_samples()),
                make_metadata_item(audio.get_metadata()),
            ])

        sensors_item.addChildren([
            QTreeWidgetItem(["accelerometer"]),
            QTreeWidgetItem(["ambient_temperature"]),
            audio_item,
            QTreeWidgetItem(["compressed_audio"]),
            QTreeWidgetItem(["gravity"]),
            QTreeWidgetItem(["gyroscope"]),
            QTreeWidgetItem(["image"]),
            QTreeWidgetItem(["light"]),
            QTreeWidgetItem(["linear_acceleration"]),
            QTreeWidgetItem(["location"]),
            QTreeWidgetItem(["magnetometer"]),
            QTreeWidgetItem(["orientation"]),
            QTreeWidgetItem(["pressure"]),
            QTreeWidgetItem(["proximity"]),
            QTreeWidgetItem(["relative_humidity"]),
            QTreeWidgetItem(["rotation_vector"]),
            make_metadata_item(sensors.get_metadata()),
        ])

        self.currentItemChanged.connect(self.check_if_value_item)

        self.insertTopLevelItem(0, QTreeWidgetItem(["api", str(wrapped_packet.get_api())]))
        self.insertTopLevelItem(1, station_info_item)
        self.insertTopLevelItem(2, timing_info_item)
        self.insertTopLevelItem(3, sensors_item)
        self.insertTopLevelItem(4, make_metadata_item(wrapped_packet.get_metadata()))
        self.setHeaderLabels(["Field", "Value"])
        self.setColumnWidth(0, 400)



"""
This module provides low level aggregate read functionality for RedVox API M data.
"""

from collections import defaultdict
from datetime import datetime, timedelta
from dataclasses import dataclass
from functools import reduce
from glob import glob
from multiprocessing import Pool
import os.path
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Set

from redvox.common.date_time_utils import datetime_from_epoch_microseconds_utc as dt_us
from redvox.api1000.common.lz4 import decompress
import redvox.api1000.proto.redvox_api_m_pb2 as pb
from redvox.api1000.wrapped_redvox_packet.station_information import OsType
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM


# noinspection DuplicatedCode
@dataclass
class StationSummary:
    """
    Contains a summary of each stations data read result.
    """
    station_id: str
    station_uuid: str
    auth_id: str
    os: OsType
    os_version: str
    app_version: str
    audio_sampling_rate: float
    total_packets: int
    total_duration: timedelta
    start_dt: datetime
    end_dt: datetime

    @staticmethod
    def from_packets(packets: List[WrappedRedvoxPacketM]) -> 'StationSummary':
        first_packet: WrappedRedvoxPacketM = packets[0]
        last_packet: WrappedRedvoxPacketM = packets[-1]
        total_duration: timedelta = reduce(lambda acc, packet: acc + packet.get_packet_duration(),
                                           packets,
                                           timedelta(seconds=0.0))
        start_dt: datetime = dt_us(first_packet.get_timing_information().get_packet_start_mach_timestamp())
        end_dt: datetime = dt_us(last_packet.get_timing_information().get_packet_start_mach_timestamp()) + \
                           last_packet.get_packet_duration()

        station_info = first_packet.get_station_information()
        audio = first_packet.get_sensors().get_audio()
        return StationSummary(
            station_info.get_id(),
            station_info.get_uuid(),
            station_info.get_auth_id(),
            station_info.get_os(),
            station_info.get_os_version(),
            station_info.get_app_version(),
            audio.get_sample_rate() if audio is not None else float("NaN"),
            len(packets),
            total_duration,
            start_dt,
            end_dt
        )


@dataclass
class ReadFilter:
    """
    Filter API M files from the file system.
    """
    start_dt: Optional[datetime] = None
    end_dt: Optional[datetime] = None
    station_ids: Optional[Set[str]] = None
    extension: str = ".rdvxm"

    def with_start_dt(self, start_dt: datetime) -> 'ReadFilter':
        """
        Adds a start datetime filter.
        :param start_dt: Start datetime that files should come after.
        :return: A modified instance of this filter
        """
        self.start_dt = start_dt
        return self

    def with_start_ts(self, start_ts: float) -> 'ReadFilter':
        """
        Adds a start time filter.
        :param start_ts: Start timestamp (microseconds)
        :return: A modified instance of this filter
        """
        return self.with_start_dt(dt_us(start_ts))

    def with_end_dt(self, end_dt: datetime) -> 'ReadFilter':
        """
        Adds an end datetime filter.
        :param end_dt: Filter for which packets should come before.
        :return: A modified instance of this filter
        """
        self.end_dt = end_dt
        return self

    def with_end_ts(self, end_ts: float) -> 'ReadFilter':
        """
        Like with_end_dt, but uses a microsecond timestamp.
        :param end_ts: Timestamp microseconds.
        :return: A modified instance of this filter
        """
        return self.with_end_dt(dt_us(end_ts))

    def with_station_ids(self, station_ids: Set[str]) -> 'ReadFilter':
        """
        Add a station id filter. Filters against provided station ids.
        :param station_ids: Station ids to filter against.
        :return: A modified instance of this filter
        """
        self.station_ids = station_ids
        return self

    def with_extension(self, extension: str) -> 'ReadFilter':
        """
        Filters against a known file extension.
        :param extension: Extension to filter against
        :return: A modified instance of this filter
        """
        self.extension = extension
        return self

    def filter_dt(self, dt: datetime) -> bool:
        """
        Tests if a given datetime passes this filter.
        :param dt: Datetime to test
        :return: True if the datetime is included, False otherwise
        """
        if self.start_dt is not None and dt < self.start_dt:
            return False

        if self.end_dt is not None and dt > self.end_dt:
            return False

        return True

    def filter_path(self, path: str) -> bool:
        """
        Tests a given file system path against this filter.
        :param path: Path to test.
        :return: True if the path is accepted, False otherwise
        """
        _path: Path = Path(path)
        ext: str = "".join(_path.suffixes)
        station_ts: str = _path.stem
        split: List[str] = station_ts.split("_")
        station_id: str = split[0]
        ts: float = float(split[1])
        dt: datetime = dt_us(ts)

        if not self.filter_dt(dt):
            return False

        if self.station_ids is not None and station_id not in self.station_ids:
            return False

        if self.extension is not None and self.extension != ext:
            return False

        return True


class ReadResult:
    """
    Result of reading multiple API M files.
    """

    def __init__(self,
                 station_id_uuid_to_packets: Dict[str, List[WrappedRedvoxPacketM]]):
        """
        :param station_id_uuid_to_packets: station_id:station_uuid -> packets
        """
        self.station_id_uuid_to_packets: Dict[str, List[WrappedRedvoxPacketM]] = station_id_uuid_to_packets
        self.__station_id_to_id_uuid: Dict[str, str] = {}
        self.__station_summaries: List[StationSummary] = []

        for id_uuid, packets in self.station_id_uuid_to_packets.items():
            s: List[str] = id_uuid.split(":")
            self.__station_id_to_id_uuid[s[0]] = id_uuid
            self.__station_summaries.append(StationSummary.from_packets(packets))

    @staticmethod
    def from_packets(packets: List[WrappedRedvoxPacketM]) -> 'ReadResult':
        """
        Constructs a read result from the provided packets.
        :param packets: Packets to construct read result from.
        :return: ReadResult from provided packets
        """
        station_id_uuid_to_packets: Dict[str, List[WrappedRedvoxPacketM]] = defaultdict(list)

        for packet in packets:
            station_info = packet.get_station_information()
            station_id: str = station_info.get_id()
            station_uuid: str = station_info.get_uuid()
            id_uuid: str = f"{station_id}:{station_uuid}"
            station_id_uuid_to_packets[id_uuid].append(packet)

        return ReadResult(station_id_uuid_to_packets)

    def __get_packets_for_station_id_uuid(self, station_id_uuid) -> List[WrappedRedvoxPacketM]:
        """
        Find packets given a station_id:uuid.
        :param station_id_uuid: Station id and uuid to get packets for.
        :return: A list of wrapped packets or an empty list if none match.
        """
        if station_id_uuid in self.station_id_uuid_to_packets:
            return self.station_id_uuid_to_packets[station_id_uuid]

        return []

    def __get_packets_for_station_id(self, station_id: str) -> List[WrappedRedvoxPacketM]:
        """
        Get packets for an associated station id.
        :param station_id: The station id.
        :return: A list of wrapped packets or an empty list if none provided.
        """
        if station_id in self.__station_id_to_id_uuid:
            return self.__get_packets_for_station_id_uuid(self.__station_id_to_id_uuid[station_id])

        return []

    def get_station_summaries(self) -> List[StationSummary]:
        """
        :return: A list of StationSummaries contained in this ReadResult
        """
        return self.__station_summaries

    def get_packets_for_station_id(self, station_id: str) -> List[WrappedRedvoxPacketM]:
        """
        Gets packets either for the provided station_id or the provided station_id:uuid.
        :param station_id: station_id or station_id:uuid to get packets for.
        :return: A list of packets of an empty list of none provided.
        """
        if ":" in station_id:
            return self.__get_packets_for_station_id_uuid(station_id)
        else:
            return self.__get_packets_for_station_id(station_id)


# We need to parse the API M structured directory structure. Here, we enumerate the valid values for the various
# levels in the hierarchy.
__VALID_YEARS: Set[str] = {f"{i:04}" for i in range(2018, 2031)}
__VALID_MONTHS: Set[str] = {f"{i:02}" for i in range(1, 13)}
__VALID_DATES: Set[str] = {f"{i:02}" for i in range(1, 32)}
__VALID_HOURS: Set[str] = {f"{i:02}" for i in range(0, 24)}


def __deserialize_path(path: str):
    with open(path, "rb") as fin:
        buf: bytes = fin.read()
        debuf: bytes = decompress(buf)
        proto: pb.RedvoxPacketM = pb.RedvoxPacketM()
        proto.ParseFromString(debuf)
        return proto


def __deserialize_paths(paths: List[str], parallel: bool = False) -> List[WrappedRedvoxPacketM]:
    """
    Deserialize a list of paths into a list of WrappedRedvoxPacketMs
    :param paths: Paths to deserialize
    :param parallel: If True, a process pool is used to perform all decompression and deserialization. After
                     decompression and deserialization, wrapping takes place in the original process. If False,
                     decompression, deserialization, and wrapping take place in the original process serially.

    :return: A list of WrappedRedvoxPacketMs.
    """
    if parallel:
        pool = Pool()
        deserialized = list(pool.map(__deserialize_path, paths))
        return sorted(list(map(lambda de: WrappedRedvoxPacketM(de), deserialized)))
    else:
        return sorted(list(map(WrappedRedvoxPacketM.from_compressed_path, paths)))


def __list_subdirs(base_dir: str, valid_choices: Set[str]) -> List[str]:
    """
    Lists sub-directors in a given base directory that match the provided choices.
    :param base_dir: Base dir to find sub dirs in.
    :param valid_choices: A list of valid directory names.
    :return: A list of valid subdirs.
    """
    subdirs: Iterator[str] = map(lambda p: Path(p).name, glob(os.path.join(base_dir, "*", "")))
    return sorted(list(filter(valid_choices.__contains__, subdirs)))


def __parse_structured_layout(base_dir: str,
                              read_filter: ReadFilter = ReadFilter()) -> List[str]:
    """
    This parses a structured API M directory structure and identifies files that match the provided filter.
    :param base_dir: Base directory (should be named api1000)
    :param read_filter: Filter to filter files with
    :return: A list of wrapped packets on an empty list if none match the filter or none are found
    """
    all_paths: List[str] = []
    for year in __list_subdirs(base_dir, __VALID_YEARS):
        for month in __list_subdirs(os.path.join(base_dir, year), __VALID_MONTHS):
            for day in __list_subdirs(os.path.join(base_dir, year, month), __VALID_DATES):
                for hour in __list_subdirs(os.path.join(base_dir, year, month, day), __VALID_HOURS):
                    # Before scanning for *.rdvxm files, let's see if the current year, month, day, hour are in the
                    # filter's range. If not, we can short circuit and skip getting the *.rdvxm files.
                    if not read_filter.filter_dt(datetime(int(year),
                                                          int(month),
                                                          int(day),
                                                          int(hour))):
                        continue

                    paths: List[str] = glob(os.path.join(base_dir,
                                                         year,
                                                         month,
                                                         day,
                                                         hour,
                                                         f"*{read_filter.extension}"))
                    # Filer paths that match the predicate
                    valid_paths: List[str] = list(filter(lambda path: read_filter.filter_path(path), paths))
                    if len(valid_paths) > 0:
                        all_paths.extend(valid_paths)

    return all_paths


def read_bufs(bufs: List[bytes]) -> ReadResult:
    """
    Reads a list of API M packet buffers.
    :param bufs: Buffers to read.
    :return: A ReadResult of the read data.
    """
    wrapped_packets: List[WrappedRedvoxPacketM] = list(sorted(map(WrappedRedvoxPacketM.from_compressed_bytes, bufs)))
    return ReadResult.from_packets(wrapped_packets)


def read_structured(base_dir: str, read_filter: ReadFilter = ReadFilter()) -> ReadResult:
    """
    Read structured API M data. Structured API data is stored using the following directory hierarchy.
        api1000/[YYYY]/[MM]/[DD]/[HH]/*.rdvxm
    :param base_dir: Base directory of structured data (should be named api1000)
    :param read_filter: Filter to apply to files.
    :return: A ReadResult
    """
    paths: List[str] = __parse_structured_layout(base_dir, read_filter)
    wrapped_packets: List[WrappedRedvoxPacketM] = __deserialize_paths(paths)
    return ReadResult.from_packets(wrapped_packets)


def read_unstructured(base_dir: str, read_filter: ReadFilter = ReadFilter()) -> ReadResult:
    """
    Reads RedVox files from a provided directory.
    :param base_dir: Directory to read files from.
    :param read_filter: Filter to filter files with.
    :return: A ReadResult.
    """
    pattern: str = os.path.join(base_dir, f"*{read_filter.extension}")
    paths: List[str] = glob(os.path.join(base_dir, pattern))
    paths = list(filter(lambda path: read_filter.filter_path(path), paths))
    wrapped_packets: List[WrappedRedvoxPacketM] = __deserialize_paths(paths)
    return ReadResult.from_packets(wrapped_packets)


def stream_structured(base_dir: str, read_filter: ReadFilter = ReadFilter()) -> Iterator[WrappedRedvoxPacketM]:
    """
    Lazily loads API M data from a structured layout.
    :param base_dir: Directory to read files from.
    :param read_filter: Filter to filter files with.
    :return: An iterator that reads and loads one WrappedRedvoxPacketM at a time.
    """
    paths: List[str] = __parse_structured_layout(base_dir, read_filter)

    for path in paths:
        yield WrappedRedvoxPacketM.from_compressed_path(path)


def stream_unstructured(base_dir: str, read_filter: ReadFilter = ReadFilter()) -> Iterator[WrappedRedvoxPacketM]:
    """
    Lazily loads API M data from an unstructured layout.
    :param base_dir: Directory to read files from.
    :param read_filter: Filter to filter files with.
    :return: An iterator that reads and loads one WrappedRedvoxPacketM at a time.
    """
    pattern: str = os.path.join(base_dir, f"*{read_filter.extension}")
    paths: List[str] = glob(os.path.join(base_dir, pattern))
    paths = list(filter(lambda path: read_filter.filter_path(path), paths))

    for path in paths:
        yield WrappedRedvoxPacketM.from_compressed_path(path)



from datetime import datetime
from typing import Optional, List, Any, Union

from PySide2.QtCharts import QtCharts
from PySide2.QtCore import Qt
from PySide2.QtGui import QPainter
from PySide2.QtWidgets import QWidget, QVBoxLayout, QTableWidget, QTableWidgetItem, QHeaderView


class DateTimes:
    def __init__(self, ts_us: List[float]) -> None:
        self.ts_us: List[float] = ts_us

    @staticmethod
    def from_sr(start_ts_us: float, sample_rate: float, cnt: int) -> 'DateTimes':
        sample_interval: float = 1_000_000.0 / sample_rate
        return DateTimes([start_ts_us + sample_interval * i for i in range(cnt)])

    @staticmethod
    def from_ts_ms(ts_ms: List[float]) -> 'DateTimes':
        return DateTimes(list(map(lambda v: v * 1_000.0, ts_ms)))

    @staticmethod
    def from_ts_s(ts_s: List[float]) -> 'DateTimes':
        return DateTimes(list(map(lambda v: v * 1_000_000.0, ts_s)))

    @staticmethod
    def from_dts(dts: List[datetime]) -> 'DateTimes':
        return DateTimes(list(map(lambda dt: dt.timestamp() * 1_000_000.0, dts)))

    def ts_ms(self) -> List[float]:
        return list(map(lambda v: v / 1_000.0, self.ts_us))

    def ts_s(self) -> List[float]:
        return list(map(lambda v: v / 1_000_000.0, self.ts_us))

    def dts(self) -> List[datetime]:
        return list(map(datetime.utcfromtimestamp, self.ts_s()))


class ValuesWidget(QTableWidget):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setColumnCount(1)
        self.setHorizontalHeaderLabels(["value"])

    def update_from_values(self, values: List[Any], date_times: Optional[DateTimes] = None) -> None:
        self.clear()
        if date_times is None:
            self.setColumnCount(1)
            self.setHorizontalHeaderLabels(["values"])
        else:
            self.setColumnCount(2)
            self.setHorizontalHeaderLabels(["values", "date/time"])
            dts = date_times.dts()
        self.setRowCount(len(values))

        for i, v in enumerate(values):
            self.setItem(i, 0, QTableWidgetItem(str(v)))
            if date_times is not None:
                self.setItem(i, 1, QTableWidgetItem(str(dts[i])))

        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)


class PlotWidget(QtCharts.QChartView):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setRenderHint(QPainter.Antialiasing)

    def update_from_values(self, values: List[float], date_times: DateTimes) -> None:
        start_dt: datetime = datetime.utcfromtimestamp(date_times.ts_us[0] / 1_000_000.0)
        chart: QtCharts.QChart = QtCharts.QChart()
        chart.legend().hide()
        chart.setTitle(f"Audio Samples @ {start_dt}")

        axis_x: QtCharts.QDateTimeAxis = QtCharts.QDateTimeAxis()
        axis_x.setFormat("hh:mm:ss")
        axis_x.setTitleText("Date/Time")
        axis_x.setTickCount(5)
        chart.addAxis(axis_x, Qt.AlignBottom)

        series: QtCharts.QLineSeries = QtCharts.QLineSeries()
        ts_s: List[float] = date_times.ts_ms()
        for i, v in enumerate(values):
            series.append(ts_s[i], v)
        chart.addSeries(series)

        axis_y: QtCharts.QValueAxis = QtCharts.QValueAxis()
        axis_y.setTickCount(5)
        axis_y.setTitleText("PCM F32")
        chart.addAxis(axis_y, Qt.AlignLeft)

        series.attachAxis(axis_x)
        series.attachAxis(axis_y)

        self.setChart(chart)


class PlotAndDetailsColumn(QWidget):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.values_widget = ValuesWidget(self)
        self.plot_widget = PlotWidget(self)

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.plot_widget)
        self.layout().addWidget(self.values_widget)



class CloudApiError(Exception):
    def __init__(self, message: str = ""):
        super().__init__(message)


class AuthenticationError(CloudApiError):
    def __init__(self):
        super().__init__("Authentication Error")


class ApiConnectionError(CloudApiError):
    def __init__(self, message: str = ""):
        super().__init__(f"ApiConnectionError: {message}")



"""
This module contains functions for downloading data from AWS S3 via signed URLs.
"""

import logging
import multiprocessing
import os
from typing import Callable, List, Tuple, Optional

import numpy as np
import requests

# pylint: disable=C0103
log = logging.getLogger(__name__)


class ProcessPool:
    """
    Creates a process pool used for fetching files from S3.
    """

    def __init__(self,
                 num_processes: int,
                 func: Callable[[List[str], str, int], None],
                 data: List[str],
                 out_dir: str,
                 retries: int):
        """
        Instantiates a new process pool.
        :param num_processes: The number of processes to create.
        :param func: The function to run which should take a list of strings representing data keys and a string
                     representing the out_dir.
        :param data: The list of data keys to process.
        :param out_dir: The output directory to write the data files to.
        :param retries: The number of times the HTTP client should retry getting a file.
        """

        self.num_processes: int = num_processes
        self.func: Callable[[List[str], str, int], None] = func
        self.data: List[List[str]] = list(map(list, np.array_split(np.array(data), num_processes)))
        self.out_dir = out_dir
        self.retries = retries

    def run(self):
        """
        Runs this process pool. This will block until all processes finish fetching data.
        """
        processes: List[multiprocessing.Process] = []

        for i in range(self.num_processes):
            process: multiprocessing.Process = multiprocessing.Process(target=self.func, args=(self.data[i],
                                                                                               self.out_dir,
                                                                                               self.retries))
            processes.append(process)
            process.start()

        # Wait for all processes to finish
        for process in processes:
            process.join()


def find_between(start: str, end: str, contents: str) -> str:
    """
    Find the string contents between two other strings.
    :param start: The first string.
    :param end: The seconds string.
    :param contents: The full string.
    :return: The contents between the start and end strings.
    """
    s_idx = contents.find(start)
    e_idx = contents.find(end)
    return contents[s_idx + len(start):e_idx]


def get_file(url: str,
             retries: int,
             session: requests.Session = requests.Session()) -> Optional[bytes]:
    """
    Attempts to download a file with a configurable amount of retries.
    :param url: The url to download.
    :param retries: Number of retries.
    :param session: An instance of a session.
    :return: The bytes of the file.
    """
    try:
        resp: requests.Response = session.get(url)
        if resp.status_code == 200:
            return resp.content
        else:
            log.error("Received error response when requesting data for url=%s: %d %s",
                      url,
                      resp.status_code,
                      resp.text)
            if retries > 0:
                log.info("Retrying with %d retries", retries)
                return get_file(url, retries - 1, session)

            log.error("All retries exhausted, could not get %s", url)
            return None
    except Exception as e:
        log.error("Encountered an error while getting data for %s: %s", url, str(e))
        if retries > 0:
            log.info("Retrying with %d retries", retries)
            return get_file(url, retries - 1, session)
        log.error("All retries exhausted, could not get %s", url)
        return None


def download_file(url: str,
                  session: requests.Session,
                  out_dir: str,
                  retries: int) -> Tuple[str, int]:
    """
    Attempts to download a file from S3.
    :param url: The URL to retrieve.
    :param session: The HTTP session.
    :param out_dir: The output directory where files will be stored.
    :param retries: The number of times to retry failed file downloads.
    :return: A tuple containing the data_key and the length of the download response.
    """
    buf: Optional[bytes] = get_file(url, retries, session)

    if buf:
        data_key = find_between("/rdvxdata/", "?X-Amz-Algorithm=", url)

        directory = os.path.dirname(data_key)
        full_dir = f"{out_dir}/{directory}"
        if not os.path.exists(full_dir):
            log.info("Directory %s does not exist, creating it", full_dir)
            os.makedirs(full_dir)

        full_path = f"{out_dir}/{data_key}"
        with open(full_path, "wb") as fout:
            bytes_written = fout.write(buf)
            log.info("Wrote %s %d", full_path, bytes_written)

        return data_key, len(buf)

    return "", 0


def download_files(urls: List[str], out_dir: str, retries: int) -> None:
    """
    Download files from S3.
    :param urls: The URL of the files to download.
    :param out_dir: The output directory to store the downloaded files.
    :param retries: The number of retries for failed downloads.
    """
    session: requests.Session = requests.Session()

    for url in urls:
        data_key, resp_len = download_file(url, session, out_dir, retries)
        log.info("Recv %s with len=%d", data_key, resp_len)

    session.close()


def download_files_parallel(urls: List[str], out_dir: str, retries: int) -> None:
    """
    Distributes the file downloading across a process pool.
    :param urls: The signed URLs to download.
    :param out_dir: The output directory.
    :param retries: The number of times to retry downloading a file on failure.
    """
    process_pool: ProcessPool = ProcessPool(4,
                                            download_files,
                                            urls,
                                            out_dir,
                                            retries)
    process_pool.run()



"""
Provides a basic image gallery for API M image sensors.
"""

import sys
import datetime
from typing import Optional, TYPE_CHECKING

import numpy as np
from PySide2.QtCore import Qt, QByteArray
from PySide2.QtGui import QPixmap
from PySide2.QtWidgets import QApplication, QMainWindow, QWidget, QHBoxLayout, QLabel, QTableWidget, QTableWidgetItem, \
    QHeaderView, QSizePolicy, QAbstractItemView

import redvox

if TYPE_CHECKING:
    from redvox.api1000.wrapped_redvox_packet.sensors.image import Image


class ImageSelectionWidget(QTableWidget):
    """
    This widget provides a table view of available images in the packet.
    """

    def __init__(self,
                 image_sensor: 'Image',
                 image_view_widget: 'ImageViewWidget',
                 parent: Optional[QWidget] = None) -> None:
        super().__init__(image_sensor.get_num_images(), 2, parent=parent)

        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.resizeColumnsToContents()

        extension: str = image_sensor.get_file_ext()
        timestamps: np.ndarray = image_sensor.get_timestamps().get_timestamps()

        self.setHorizontalHeaderLabels(["File Name", "Image Sampled At"])
        for (i, ts) in enumerate(timestamps):
            name: str = f"{round(ts)}.{extension}"
            dt: datetime.datetime = datetime.datetime.utcfromtimestamp(ts / 1_000_000.0)
            self.setItem(i, 0, QTableWidgetItem(name))
            self.setItem(i, 1, QTableWidgetItem(str(dt)))

        def __update_image_at_row(row: int) -> None:
            buf: bytes = image_sensor.get_samples()[row]
            image_view_widget.set_image(buf)

        self.currentCellChanged.connect(lambda r: __update_image_at_row(r))


class ImageViewWidget(QLabel):
    """
    This widget displays and scales the image
    """

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent=None)
        self.setAlignment(Qt.AlignCenter)

    def set_image(self, buf: bytes) -> None:
        """
        Sets the image.
        :param buf: Image bytes to set
        """
        pix: QPixmap = QPixmap()
        pix.loadFromData(QByteArray(buf))
        self.setPixmap(pix.scaled(self.width(),
                                  self.height(),
                                  Qt.AspectRatioMode.KeepAspectRatio,
                                  Qt.TransformationMode.SmoothTransformation))


class ImageViewer(QWidget):
    """
    Widget that acts as the "main" widget into the application
    """

    def __init__(self, image_sensor: 'Image', parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setLayout(QHBoxLayout(self))

        image_view_widget: ImageViewWidget = ImageViewWidget()
        image_selection_widget: ImageSelectionWidget = ImageSelectionWidget(image_sensor, image_view_widget)
        size_policy: QSizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        size_policy.setHorizontalStretch(1)
        image_selection_widget.setSizePolicy(size_policy)
        self.layout().addWidget(image_selection_widget)

        size_policy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        size_policy.setHorizontalStretch(2)
        image_view_widget.setSizePolicy(size_policy)
        self.layout().addWidget(image_view_widget)


class MainWindow(QMainWindow):
    """
    Main window of the application
    """

    def __init__(self, image_sensor: 'Image', parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setWindowTitle(f"RedVox Image Viewer v{redvox.VERSION}")
        self.setCentralWidget(ImageViewer(image_sensor, self))


def start_gui(image_sensor: 'Image') -> None:
    """
    Starts the GUI from the provided image sensor.
    :param image_sensor: Image sensor
    """
    app: QApplication = QApplication(sys.argv)
    window: MainWindow = MainWindow(image_sensor)
    window.showMaximized()
    app.exec_()



from redvox.cloud.api import ApiConfig

import redvox.cloud.client as cloud_client


def main():
    client: cloud_client.CloudClient
    with cloud_client.cloud_client("redvoxcore@gmail.com",
                                   "red1Core18",
                                   ApiConfig(
                                       "http",
                                       "localhost",
                                       8080
                                   ),
                                   "The_Early_Bird_Gets_The_Worm") as client:
        print(client.request_metadata_m(
            1601424000,
            1601424000 + 300,
            ["1637610001"],
            ["metadata"]
        ).db_packets[0].metadata)


if __name__ == "__main__":
    main()



"""
This module provides utility functions for determining statistics of well structured RedVox data.
"""

from typing import Tuple, Union

import numpy as np

SAMPLE_RATE_HZ: np.ndarray = np.array([80, 800, 8000, 16000])  # list of accepted sample rates in Hz
BASE_NUMBER_POINTS: int = 4096  # the number of points to sample at the first sample rate
NUM_POINTS_FACTOR: int = 2**3  # the multiplier of points per increased sample rate

# total multiplier of base number of points, 1 multiplier per sample rate
POINTS_FACTOR_ARRAY: np.ndarray = np.array([1, NUM_POINTS_FACTOR, NUM_POINTS_FACTOR**2, 2*NUM_POINTS_FACTOR**2])

# total number of points per sample rate
DURATION_TOTAL_POINTS: np.ndarray = np.array(POINTS_FACTOR_ARRAY * BASE_NUMBER_POINTS)

# expected duration of packets in seconds
DURATION_SECONDS: np.ndarray = np.divide(DURATION_TOTAL_POINTS, SAMPLE_RATE_HZ)


def get_file_stats(sample_rate: Union[float, int]) -> Tuple[int, float]:
    """
    Get the number of samples in a decoder file and its duration in seconds.
    :param sample_rate: int or float, sample rate
    :returns: number of samples in file as int and file time duration in seconds as float
    """
    try:
        position: int = np.where(SAMPLE_RATE_HZ == sample_rate)[0][0]
    except Exception:
        raise ValueError(f'Sample rate {sample_rate} for mic data not recognized.')

    return DURATION_TOTAL_POINTS[position], DURATION_SECONDS[position]


def get_num_points_from_sample_rate(sample_rate: Union[float, int]) -> int:
    """
    Returns the number of data points in a packet given a sample rate
    :param sample_rate: A valid sample rate from the constants above
    :return: the number of data points in the packet in seconds
    """
    try:
        position: int = np.where(SAMPLE_RATE_HZ == sample_rate)[0][0]
        return DURATION_TOTAL_POINTS[position]
    except Exception:
        raise ValueError(f"Unknown sample rate {sample_rate} given to compute number of data points!")


def get_duration_seconds_from_sample_rate(sample_rate: Union[float, int]) -> float:
    """
    Returns the duration of a packet in seconds given a sample rate
    :param sample_rate: A valid sample rate from the constants above
    :return: the duration of the packet in seconds
    """
    try:
        position: int = np.where(SAMPLE_RATE_HZ == sample_rate)[0][0]
        return DURATION_SECONDS[position]
    except Exception:
        raise ValueError(f"Unknown sample rate {sample_rate} given to compute duration!")



"""
This module provides a command line interface (CLI) for converting, viewing, and downloading RedVox data files.
"""

import argparse
import logging
import os.path
import sys
from typing import Dict, List, Optional

from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM

from redvox.api1000.gui.image_viewer import start_gui
from redvox.api1000.wrapped_redvox_packet.sensors.image import Image, ImageCodec

import redvox.cloud.api as cloud_api
import redvox.cloud.client as cloud_client
import redvox.cloud.data_api as data_api
import redvox.cli.conversions as conversions
import redvox.cli.data_req as data_req

# pylint: disable=C0103
log = logging.getLogger(__name__)


def check_path(path: str, path_is_file: bool = True, file_ext: Optional[str] = None) -> bool:
    """
    Checks that the passed in path exists.
    :param file_ext: Optional extension to check.
    :param path: The path to check.
    :param path_is_file: The path is a file when True or a directory when False.
    :return: True if the path exists, False otherwise.
    """
    if path_is_file:
        return os.path.isfile(path) and (file_ext is None or os.path.basename(path).endswith(file_ext))
    else:
        return os.path.isdir(path)


def check_files(paths: List[str], file_ext: Optional[str] = None) -> bool:
    """
    Checks this given files to determine if they exist.
    :param paths: The paths to check.
    :param file_ext: An optional file extension to filter against.
    :return: True if all paths exist, False otherwise
    """
    invalid_paths: List[str] = list(filter(lambda path: not check_path(path, file_ext=file_ext), paths))
    if len(invalid_paths) > 0:
        log.error("%d invalid paths found", len(invalid_paths))
        for invalid_path in invalid_paths:
            log.error("Invalid path %s", invalid_path)
        return False
    return True


def check_out_dir(out_dir: Optional[str] = None) -> bool:
    """
    Checks if a given directory exists.
    :param out_dir: The directory to check.
    :return: True if it exists, False otherwise.
    """
    if out_dir is not None and not check_path(out_dir, path_is_file=False):
        log.error("out_dir is invalid: %s", out_dir)
        return False
    return True


def determine_exit(status: bool) -> None:
    """
    Determine the exit status and exit the CLI.
    :param status: True will exit with a status of 0 and False will exit with a status of 1.
    """
    if status:
        log.info("Exiting with status = 0")
        sys.exit(0)

    log.error("Exiting with status = 1")
    sys.exit(1)


def rdvxz_to_rdvxm(args) -> None:
    """
    Convert rdvxz to rdvxm
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxz_paths, ".rdvxz"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.rdvxz_to_rdvxm(args.rdvxz_paths, args.out_dir))


def rdvxm_to_rdvxz(args) -> None:
    """
    Convert rdvxm to rdvxz
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxm_paths, ".rdvxm"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.rdvxm_to_rdvxz(args.rdvxm_paths, args.out_dir))


def rdvxz_to_json_args(args) -> None:
    """
    Wrapper function that calls the to_json conversion.
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxz_paths, ".rdvxz"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.rdvxz_to_json(args.rdvxz_paths, args.out_dir))


def rdvxm_to_json_args(args) -> None:
    """
    Wrapper function that calls the to_json conversion.
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxm_paths, ".rdvxm"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.rdvxm_to_json(args.rdvxm_paths, args.out_dir))


def json_to_rdvxz_args(args) -> None:
    """
    Wrapper function that calls the to_rdvxz conversion.
    :param args: Args from argparse.
    """
    if not check_files(args.json_paths, ".json"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.json_to_rdvxz(args.json_paths, args.out_dir))


def json_to_rdvxm_args(args) -> None:
    """
    Wrapper function that calls the to_rdvxm conversion.
    :param args: Args from argparse.
    """
    if not check_files(args.json_paths, ".json"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.json_to_rdvxm(args.json_paths, args.out_dir))


def rdvxz_print_stdout_args(args) -> None:
    """
    Wrapper function that calls the print to stdout.
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxz_paths, ".rdvxz"):
        determine_exit(False)

    determine_exit(conversions.rdvxz_print_stdout(args.rdvxz_paths))


def rdvxm_print_stdout_args(args) -> None:
    """
    Wrapper function that calls the print to stdout.
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxm_paths, ".rdvxm"):
        determine_exit(False)

    determine_exit(conversions.rdvxm_print_stdout(args.rdvxm_paths))


def validate_rdvxm_args(args) -> None:
    """
    Validates the args
    :param args: Args from argparse
    """
    if not check_files(args.rdvxm_paths, ".rdvxm"):
        determine_exit(False)

    determine_exit(conversions.validate_rdvxm(args.rdvxm_paths))


def data_req_args(args) -> None:
    """
    Wrapper function that calls the data_req.
    :param args: Args from argparse.
    """
    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(data_req.make_data_req(args.out_dir,
                                          args.protocol,
                                          args.host,
                                          args.port,
                                          args.email,
                                          args.password,
                                          args.req_start_s,
                                          args.req_end_s,
                                          args.redvox_ids,
                                          args.retries,
                                          args.secret_token))


def data_req_report(protocol: str,
                    host: str,
                    port: int,
                    email: str,
                    password: str,
                    report_id: str,
                    out_dir: str,
                    retries: int,
                    secret_token: Optional[str] = None) -> bool:
    """
    Uses the built-in cloud based HTTP API to generate a signed URL for report data and then downloads the report data.
    :param protocol: Either http or https.
    :param host: The data service host.
    :param port: The data service port.
    :param email: The email of the RedVox user.
    :param password: The password of the RedVox user.
    :param report_id: The full RedVox report id.
    :param out_dir: The output directory to play the report distribution.
    :param retries: Number of times to attempt to retry the download on failed attempts.
    :param secret_token: The shared secret if utilized by the API server.
    """
    api_config: cloud_api.ApiConfig = cloud_api.ApiConfig(protocol, host, port)
    client = cloud_client.CloudClient(email,
                                      password,
                                      api_conf=api_config,
                                      secret_token=secret_token)
    resp: Optional[data_api.ReportDataResp] = client.request_report_data(report_id)
    client.close()

    if resp:
        resp.download_fs(out_dir, retries)
        return True

    log.error("Response was None")
    return False


def data_req_report_args(args) -> None:
    """
    Wrapper function that calls the data_req_report.
    :param args: Args from argparse.
    """
    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(data_req_report(args.protocol,
                                   args.host,
                                   args.port,
                                   args.email,
                                   args.password,
                                   args.report_id,
                                   args.out_dir,
                                   args.retries,
                                   args.secret_token))


def gallery(rdvxm_paths: List[str]) -> bool:
    """
    Displays a gallery of images from the combined images collected from the given paths.
    :param rdvxm_paths: Paths to collect images from.
    :return: True if this completes successfully, False otherwise
    """
    # Create a new image sensor to hold images from all packets
    image: Image = Image.new()
    # noinspection PyTypeChecker
    image.set_image_codec(ImageCodec.JPG)

    packets: List[WrappedRedvoxPacketM] = list(map(WrappedRedvoxPacketM.from_compressed_path, rdvxm_paths))

    for packet in packets:
        image_sensor: Optional[Image] = packet.get_sensors().get_image()
        if image_sensor is not None:
            image.get_timestamps().append_timestamps(image_sensor.get_timestamps().get_timestamps())
            image.append_values(image_sensor.get_samples())

    start_gui(image)
    return True


def gallery_args(args) -> None:
    """
    CLI function for opening image gallery.
    :param args: Args for passing to gallery function.
    """
    determine_exit(gallery(args.rdvxm_paths))


def main():
    """
    Entry point into the CLI.
    """
    parser: argparse.ArgumentParser = argparse.ArgumentParser("redvox-cli",
                                                              description="Command line tools for viewing, converting,"
                                                                          " and downloading RedVox data.")
    parser.add_argument("--verbose",
                        "-v",
                        help="Enable verbose logging",
                        action="count",
                        default=0)

    sub_parser = parser.add_subparsers()
    sub_parser.required = True
    sub_parser.dest = "command"

    # Gallery
    gallery_parser = sub_parser.add_parser("gallery")
    gallery_parser.add_argument("rdvxm_paths",
                                help="One or more rdvxm files",
                                nargs="+")
    gallery_parser.set_defaults(func=gallery_args)

    # rdvxz -> rdvxm
    rdvxz_to_rdvxm_parser = sub_parser.add_parser("rdvxz_to_rdvxm",
                                                  help="Convert rdvxz (API 900) to rdvxm (API 1000/M) files")
    rdvxz_to_rdvxm_parser.add_argument("rdvxz_paths",
                                       help="One or more rdvxz files to convert to json files",
                                       nargs="+")
    rdvxz_to_rdvxm_parser.add_argument("--out_dir",
                                       "-o",
                                       help="Optional output directory (will use same directory as source files by "
                                            "default)")
    rdvxz_to_rdvxm_parser.set_defaults(func=rdvxz_to_rdvxm)

    # rdvxm -> rdvxz
    rdvxm_to_rdvxz_parser = sub_parser.add_parser("rdvxm_to_rdvxz",
                                                  help="Convert rdvxm (API 1000/M) to rdvxx (API 900) files")
    rdvxm_to_rdvxz_parser.add_argument("rdvxm_paths",
                                       help="One or more rdvxm files to convert to json files",
                                       nargs="+")
    rdvxm_to_rdvxz_parser.add_argument("--out_dir",
                                       "-o",
                                       help="Optional output directory (will use same directory as source files by "
                                            "default)")
    rdvxm_to_rdvxz_parser.set_defaults(func=rdvxm_to_rdvxz)

    # rdvxz -> json
    rdvxz_to_json_parser = sub_parser.add_parser("rdvxz_to_json",
                                                 help="Convert rdvxz files to json files")
    rdvxz_to_json_parser.add_argument("rdvxz_paths",
                                      help="One or more rdvxz files to convert to json files",
                                      nargs="+")
    rdvxz_to_json_parser.add_argument("--out_dir",
                                      "-o",
                                      help="Optional output directory (will use same directory as source files by "
                                           "default)")
    rdvxz_to_json_parser.set_defaults(func=rdvxz_to_json_args)

    rdvxm_to_json_parser = sub_parser.add_parser("rdvxm_to_json",
                                                 help="Convert rdvxm files to json files")
    rdvxm_to_json_parser.add_argument("rdvxm_paths",
                                      help="One or more rdvxm files to convert to json files",
                                      nargs="+")
    rdvxm_to_json_parser.add_argument("--out_dir",
                                      "-o",
                                      help="Optional output directory (will use same directory as source files by "
                                           "default)")
    rdvxm_to_json_parser.set_defaults(func=rdvxm_to_json_args)

    # json -> rdvxz
    json_to_rdvxz_parser = sub_parser.add_parser("json_to_rdvxz",
                                                 help="Convert json files to rdvxz files")
    json_to_rdvxz_parser.add_argument("json_paths",
                                      help="One or more json files to convert to rdvxz files",
                                      nargs="+")
    json_to_rdvxz_parser.add_argument("--out_dir",
                                      "-o",
                                      help="Optional output directory (will use same directory as source files by "
                                           "default)")
    json_to_rdvxz_parser.set_defaults(func=json_to_rdvxz_args)

    json_to_rdvxm_parser = sub_parser.add_parser("json_to_rdvxm",
                                                 help="Convert json files to rdvxm files")
    json_to_rdvxm_parser.add_argument("json_paths",
                                      help="One or more json files to convert to rdvxm files",
                                      nargs="+")
    json_to_rdvxm_parser.add_argument("--out_dir",
                                      "-o",
                                      help="Optional output directory (will use same directory as source files by "
                                           "default)")
    json_to_rdvxm_parser.set_defaults(func=json_to_rdvxm_args)

    # print
    rdvxz_print_parser = sub_parser.add_parser("printz",
                                               help="Print contents of rdvxz files to stdout")
    rdvxz_print_parser.add_argument("rdvxz_paths",
                                    help="One or more rdvxz files to print",
                                    nargs="+")
    rdvxz_print_parser.set_defaults(func=rdvxz_print_stdout_args)

    rdvxm_print_parser = sub_parser.add_parser("printm",
                                               help="Print contents of rdvxm files to stdout")
    rdvxm_print_parser.add_argument("rdvxm_paths",
                                    help="One or more rdvxm files to print",
                                    nargs="+")
    rdvxm_print_parser.set_defaults(func=rdvxm_print_stdout_args)

    # validation
    rdvxm_validation_parser = sub_parser.add_parser("validatem",
                                                    help="Validate the structure of API M files")
    rdvxm_validation_parser.add_argument("rdvxm_paths",
                                         help="One or more rdvxm files to print",
                                         nargs="+")
    rdvxm_validation_parser.set_defaults(func=validate_rdvxm_args)

    # data_req
    data_req_parser = sub_parser.add_parser("data_req",
                                            help="Request bulk RedVox data from the RedVox servers")
    data_req_parser.add_argument("--out_dir",
                                 "-o",
                                 help="The output directory that RedVox files will be written to (default=.)",
                                 default=".")
    data_req_parser.add_argument("--retries",
                                 "-r",
                                 help="The number of times the client should retry getting a file on failure "
                                      "(default=1)",
                                 default=1,
                                 choices=set(range(0, 6)),
                                 type=int)
    data_req_parser.add_argument("--host",
                                 "-H",
                                 help="Data server host (default=redvox.io)",
                                 default="redvox.io")
    data_req_parser.add_argument("--port",
                                 "-p",
                                 type=int,
                                 help="Data server port (default=8080)",
                                 default=8080)
    data_req_parser.add_argument("--protocol",
                                 help="One of either http or https (default https)",
                                 choices=["https", "http"],
                                 default="https")
    data_req_parser.add_argument("--secret_token",
                                 help="A shared secret token provided by RedVox required for accessing the data "
                                      "request service")
    data_req_parser.add_argument("email",
                                 help="redvox.io account email")
    data_req_parser.add_argument("password",
                                 help="redvox.io account password")
    data_req_parser.add_argument("req_start_s",
                                 type=int,
                                 help="Data request start as number of seconds since the epoch UTC")
    data_req_parser.add_argument("req_end_s",
                                 type=int,
                                 help="Data request end as number of seconds since the epoch UTC")
    data_req_parser.add_argument("redvox_ids",
                                 nargs="+",
                                 help="A list of RedVox ids delimited by a space")
    data_req_parser.set_defaults(func=data_req_args)

    # data req report
    data_req_report_parser = sub_parser.add_parser("data_req_report",
                                                   help="Request bulk RedVox data from the RedVox servers")
    data_req_report_parser.add_argument("--out_dir",
                                        "-o",
                                        help="The output directory that RedVox files will be written to (default=.)",
                                        default=".")
    data_req_report_parser.add_argument("--retries",
                                        "-r",
                                        help="The number of times the client should retry getting a file on failure "
                                             "(default=1)",
                                        default=1,
                                        choices=set(range(0, 6)),
                                        type=int)
    data_req_report_parser.add_argument("--host",
                                        "-H",
                                        help="Data server host (default=redvox.io)",
                                        default="redvox.io")
    data_req_report_parser.add_argument("--port",
                                        "-p",
                                        type=int,
                                        help="Data server port (default=8080)",
                                        default=8080)
    data_req_report_parser.add_argument("--protocol",
                                        help="One of either http or https (default https)",
                                        choices=["https", "http"],
                                        default="https")
    data_req_report_parser.add_argument("--secret_token",
                                        help="A shared secret token provided by RedVox required for accessing the data "
                                             "request service")
    data_req_report_parser.add_argument("email",
                                        help="redvox.io account email")
    data_req_report_parser.add_argument("password",
                                        help="redvox.io account password")
    data_req_report_parser.add_argument("report_id",
                                        type=str,
                                        help="The full report id that data is being requested for")
    data_req_report_parser.set_defaults(func=data_req_report_args)

    # Parse the args
    args = parser.parse_args()

    # Setup logging
    log_levels: Dict[int, str] = {
        0: "WARN",
        1: "INFO",
        2: "DEBUG"
    }
    log_level: str = log_levels[args.verbose] if args.verbose in log_levels else log_levels[0]
    logging.basicConfig(level=log_level,
                        format="[%(levelname)s:%(process)d:%(filename)s:%(module)s:%(funcName)s:%(lineno)d:%(asctime)s]"
                               " %(message)s")

    log.info("Running with args=%s and log_level=%s",
             str(args),
             log_level)

    # Try calling the appropriate handler
    # pylint: disable=W0703
    try:
        args.func(args)
    except Exception as e:
        log.error("Encountered an error: %s", str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()
